import decimal
import json
import logging
import re
from collections import defaultdict
from copy import deepcopy
from io import BytesIO
from typing import Dict, Optional, Set, Tuple, Union, cast

from fastavro import repository, validate
from fastavro.schema import load_schema

from confluent_kafka.schema_registry.serde import FieldTransform, FieldType, RuleConditionError, RuleContext

from .schema_registry_client import RuleKind, Schema

__all__ = [
    'AvroMessage',
    'AvroSchema',
    '_schema_loads',
    'LocalSchemaRepository',
    'parse_schema_with_repo',
    'transform',
    '_transform_field',
    'get_type',
    '_disjoint',
    '_resolve_union',
    'get_inline_tags',
    '_get_inline_tags_recursively',
    '_implied_namespace',
]

AVRO_TYPE = "AVRO"

AvroMessage = Union[
    None,  # 'null' Avro type
    str,  # 'string' and 'enum'
    float,  # 'float' and 'double'
    int,  # 'int' and 'long'
    decimal.Decimal,  # 'fixed'
    bool,  # 'boolean'
    bytes,  # 'bytes'
    list,  # 'array'
    dict,  # 'map' and 'record'
    tuple,  # wrapped union type
]
AvroSchema = Union[str, list, dict]

log = logging.getLogger(__name__)


class _ContextStringIO(BytesIO):
    """
    Wrapper to allow use of StringIO via 'with' constructs.
    """

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()
        return False


def _schema_loads(schema_str: str) -> Schema:
    """
    Instantiate a Schema instance from a declaration string.

    Args:
        schema_str (str): Avro Schema declaration.

    .. _Schema declaration:
        https://avro.apache.org/docs/current/spec.html#schemas

    Returns:
        Schema: A Schema instance.
    """

    schema_str = schema_str.strip()

    # canonical form primitive declarations are not supported
    if schema_str[0] != "{" and schema_str[0] != "[":
        schema_str = '{"type":' + schema_str + '}'

    return Schema(schema_str, schema_type='AVRO')


class LocalSchemaRepository(repository.AbstractSchemaRepository):
    def __init__(self, schemas):
        self.schemas = schemas

    def load(self, subject):
        return self.schemas.get(subject)


def parse_schema_with_repo(schema_str: str, named_schemas: Dict[str, AvroSchema]) -> AvroSchema:
    copy = deepcopy(named_schemas)
    copy["$root"] = json.loads(schema_str)
    repo = LocalSchemaRepository(copy)
    return load_schema("$root", repo=repo)


def transform(
    ctx: RuleContext, schema: AvroSchema, message: AvroMessage, field_transform: FieldTransform
) -> AvroMessage:
    if message is None or schema is None:
        return message
    field_ctx = ctx.current_field()
    if field_ctx is not None:
        field_ctx.field_type = get_type(schema)
    if isinstance(schema, list):
        subschema, submessage = _resolve_union(schema, message)
        if subschema is None:
            return message
        submessage = transform(ctx, subschema, submessage, field_transform)
        if isinstance(message, tuple) and len(message) == 2:
            return (message[0], submessage)
        return submessage
    elif isinstance(schema, dict):
        schema_type = schema.get("type")
        if schema_type == 'array':
            if not isinstance(message, list):
                log.warning("Incompatible message type for array schema")
                return message
            return [transform(ctx, schema["items"], item, field_transform) for item in message]
        elif schema_type == 'map':
            if not isinstance(message, dict):
                log.warning("Incompatible message type for map schema")
                return message
            return {key: transform(ctx, schema["values"], value, field_transform) for key, value in message.items()}
        elif schema_type == 'record':
            if not isinstance(message, dict):
                log.warning("Incompatible message type for record schema")
                return message
            fields = schema["fields"]
            for field in fields:
                if field["name"] not in message:
                    continue
                _transform_field(ctx, schema, field, message, field_transform)
            return message

    if field_ctx is not None:
        rule_tags = ctx.rule.tags
        if not rule_tags or not _disjoint(set(rule_tags), field_ctx.tags):
            return field_transform(ctx, field_ctx, message)
    return message


def _transform_field(ctx: RuleContext, schema: dict, field: dict, message: dict, field_transform: FieldTransform):
    field_type = field["type"]
    name = field["name"]
    full_name = schema["name"] + "." + name
    try:
        ctx.enter_field(message, full_name, name, get_type(field_type), None)
        value = message[name]
        new_value = transform(ctx, field_type, value, field_transform)
        if ctx.rule.kind == RuleKind.CONDITION:
            if new_value is False:
                raise RuleConditionError(ctx.rule)
        else:
            message[name] = new_value
    finally:
        ctx.exit_field()


def get_type(schema: AvroSchema) -> FieldType:
    if isinstance(schema, list):
        return FieldType.COMBINED
    elif isinstance(schema, dict):
        schema_type = schema.get("type")
    else:
        # string schemas; this could be either a named schema or a primitive type
        schema_type = schema

    if schema_type == 'record':
        return FieldType.RECORD
    elif schema_type == 'enum':
        return FieldType.ENUM
    elif schema_type == 'array':
        return FieldType.ARRAY
    elif schema_type == 'map':
        return FieldType.MAP
    elif schema_type == 'union':
        return FieldType.COMBINED
    elif schema_type == 'fixed':
        return FieldType.FIXED
    elif schema_type == 'string':
        return FieldType.STRING
    elif schema_type == 'bytes':
        return FieldType.BYTES
    elif schema_type == 'int':
        return FieldType.INT
    elif schema_type == 'long':
        return FieldType.LONG
    elif schema_type == 'float':
        return FieldType.FLOAT
    elif schema_type == 'double':
        return FieldType.DOUBLE
    elif schema_type == 'boolean':
        return FieldType.BOOLEAN
    elif schema_type == 'null':
        return FieldType.NULL
    else:
        return FieldType.NULL


def _disjoint(tags1: Set[str], tags2: Set[str]) -> bool:
    for tag in tags1:
        if tag in tags2:
            return False
    return True


def _resolve_union(schema: AvroSchema, message: AvroMessage) -> Tuple[Optional[AvroSchema], AvroMessage]:
    is_wrapped_union = isinstance(message, tuple) and len(message) == 2
    is_typed_union = isinstance(message, dict) and '-type' in message
    for subschema in schema:
        try:
            if is_wrapped_union:
                if isinstance(subschema, dict):
                    dict_schema = cast(dict, subschema)
                    tuple_message = cast(tuple, message)
                    if dict_schema["name"] == tuple_message[0]:
                        return (dict_schema, tuple_message[1])
            elif is_typed_union:
                if isinstance(subschema, dict):
                    dict_schema = cast(dict, subschema)
                    dict_message = cast(dict, message)
                    if dict_schema["name"] == dict_message['-type']:
                        return (dict_schema, dict_message)
            else:
                validate(message, subschema)
                return (subschema, message)
        except:  # noqa: E722
            continue
    return (None, message)


def get_inline_tags(schema: AvroSchema) -> Dict[str, Set[str]]:
    inline_tags: Dict[str, Set[str]] = defaultdict(set)
    _get_inline_tags_recursively('', '', schema, inline_tags)
    return inline_tags


def _get_inline_tags_recursively(ns: str, name: str, schema: Optional[AvroSchema], tags: Dict[str, Set[str]]):
    if schema is None:
        return
    if isinstance(schema, list):
        for subschema in schema:
            _get_inline_tags_recursively(ns, name, subschema, tags)
    elif not isinstance(schema, dict):
        # string schemas; this could be either a named schema or a primitive type
        return
    else:
        schema_type = schema.get("type")
        if schema_type == 'array':
            _get_inline_tags_recursively(ns, name, schema.get("items"), tags)
        elif schema_type == 'map':
            _get_inline_tags_recursively(ns, name, schema.get("values"), tags)
        elif schema_type == 'record':
            record_ns = schema.get("namespace")
            record_name = schema.get("name")
            if record_ns is None:
                record_ns = _implied_namespace(name)
            if record_ns is None:
                record_ns = ns
            # Ensure record_name is not None and doesn't already have namespace prefix
            if record_name is not None and record_ns != '' and not record_name.startswith(record_ns):
                record_name = f"{record_ns}.{record_name}"
            fields = schema["fields"]
            for field in fields:
                field_tags = field.get("confluent:tags")
                field_name = field.get("name")
                field_type = field.get("type")
                # Ensure all required fields are present before building tag key
                if field_tags is not None and field_name is not None and record_name is not None:
                    tags[record_name + '.' + field_name].update(field_tags)
                if field_type is not None and record_name is not None:
                    _get_inline_tags_recursively(record_ns, record_name, field_type, tags)


def _implied_namespace(name: str) -> Optional[str]:
    match = re.match(r"^(.*)\.[^.]+$", name)
    return match.group(1) if match else None



# Copyright 2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Type, Union


class ConversionUtil:
    @staticmethod
    def convert_to_enum(val: Union[str, int, Enum], enum_clazz: Type[Enum]) -> Enum:
        if type(enum_clazz) is not type(Enum):
            raise TypeError("'enum_clazz' must be of type Enum")

        if isinstance(val, str):
            # Allow it to be specified as case-insensitive string, for convenience.
            try:
                val = enum_clazz[val.upper()]
            except KeyError:
                raise ValueError("Unknown value \"%s\": should be a %s" % (val, enum_clazz.__name__))

        elif isinstance(val, int):
            # The C-code passes restype as an int, convert to enum.
            val = enum_clazz(val)

        elif not isinstance(val, enum_clazz):
            raise TypeError("Unknown value \"%s\": should be a %s" % (val, enum_clazz.__name__))

        return val



# Copyright 2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List

from ..cimpl import KafkaError

try:
    string_type = basestring  # type: ignore[name-defined]
except NameError:
    string_type = str


class ValidationUtil:
    @staticmethod
    def check_multiple_not_none(obj: Any, vars_to_check: List[str]) -> None:
        for param in vars_to_check:
            ValidationUtil.check_not_none(obj, param)

    @staticmethod
    def check_not_none(obj: Any, param: str) -> None:
        if getattr(obj, param) is None:
            raise ValueError("Expected %s to be not None" % (param,))

    @staticmethod
    def check_multiple_is_string(obj: Any, vars_to_check: List[str]) -> None:
        for param in vars_to_check:
            ValidationUtil.check_is_string(obj, param)

    @staticmethod
    def check_is_string(obj: Any, param: str) -> None:
        param_value = getattr(obj, param)
        if param_value is not None and not isinstance(param_value, string_type):
            raise TypeError("Expected %s to be a string" % (param,))

    @staticmethod
    def check_kafka_errors(errors: List[KafkaError]) -> None:
        if not isinstance(errors, list):
            raise TypeError("errors should be None or a list")
        for error in errors:
            if not isinstance(error, KafkaError):
                raise TypeError("Expected list of KafkaError")

    @staticmethod
    def check_kafka_error(error: KafkaError) -> None:
        if not isinstance(error, KafkaError):
            raise TypeError("Expected error to be a KafkaError")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from ._model import Node  # noqa: F401
from ._model import (
    ConsumerGroupState,
    ConsumerGroupTopicPartitions,
    ConsumerGroupType,
    ElectionType,
    IsolationLevel,
    TopicCollection,
    TopicPartitionInfo,
)
from .cimpl import (
    OFFSET_BEGINNING,
    OFFSET_END,
    OFFSET_INVALID,
    OFFSET_STORED,
    TIMESTAMP_CREATE_TIME,
    TIMESTAMP_LOG_APPEND_TIME,
    TIMESTAMP_NOT_AVAILABLE,
    Consumer,
    Message,
    Producer,
    TopicPartition,
    Uuid,
    consistent,
    fnv1a,
    libversion,
    murmur2,
    version,
)
from .deserializing_consumer import DeserializingConsumer
from .error import KafkaError, KafkaException
from .serializing_producer import SerializingProducer

__all__ = [
    "admin",
    "Consumer",
    "aio",
    "KafkaError",
    "KafkaException",
    "kafkatest",
    "libversion",
    "version",
    "murmur2",
    "consistent",
    "fnv1a",
    "Message",
    "OFFSET_BEGINNING",
    "OFFSET_END",
    "OFFSET_INVALID",
    "OFFSET_STORED",
    "Producer",
    "DeserializingConsumer",
    "SerializingProducer",
    "TIMESTAMP_CREATE_TIME",
    "TIMESTAMP_LOG_APPEND_TIME",
    "TIMESTAMP_NOT_AVAILABLE",
    "TopicPartition",
    "Node",
    "ConsumerGroupTopicPartitions",
    "ConsumerGroupState",
    "ConsumerGroupType",
    "Uuid",
    "IsolationLevel",
    "TopicCollection",
    "TopicPartitionInfo",
    "ElectionType",
]


__version__ = version()


class ThrottleEvent(object):
    """
    ThrottleEvent contains details about a throttled request.
    Set up a throttle callback by setting the ``throttle_cb`` configuration
    property to a callable that takes a ThrottleEvent object as its only argument.
    The callback will be triggered from poll(), consume() or flush() when a request
    has been throttled by the broker.

    This class is typically not user instantiated.

    :ivar str broker_name: The hostname of the broker which throttled the request
    :ivar int broker_id: The broker id
    :ivar float throttle_time: The amount of time (in seconds) the broker throttled (delayed) the request
    """

    def __init__(self, broker_name: str, broker_id: int, throttle_time: float) -> None:
        self.broker_name = broker_name
        self.broker_id = broker_id
        self.throttle_time = throttle_time

    def __str__(self) -> str:
        return "{}/{} throttled for {} ms".format(self.broker_name, self.broker_id, int(self.throttle_time * 1000))


def _resolve_plugins(plugins: str) -> str:
    """Resolve embedded plugins from the wheel's library directory.

    For internal module use only.

    :param str plugins: The plugin.library.paths value
    """
    from sys import platform

    # Location of __init__.py and the embedded library directory
    basedir = os.path.dirname(__file__)

    if platform in ("win32", "cygwin"):
        paths_sep = ";"
        ext = ".dll"
        libdir = basedir
    elif platform in ("linux", "linux2"):
        paths_sep = ":"
        ext = ".so"
        libdir = os.path.join(basedir, ".libs")
    elif platform == "darwin":
        paths_sep = ":"
        ext = ".dylib"
        libdir = os.path.join(basedir, ".dylibs")
    else:
        # Unknown platform, there are probably no embedded plugins.
        return plugins

    if not os.path.isdir(libdir):
        # No embedded library directory, probably not a wheel installation.
        return plugins

    resolved = []
    for plugin in plugins.split(paths_sep):
        if "/" in plugin or "\\" in plugin:
            # Path specified, leave unchanged
            resolved.append(plugin)
            continue

        # See if the plugin can be found in the wheel's
        # embedded library directory.
        # The user might not have supplied a file extension, so try both.
        good = None
        for file in [plugin, plugin + ext]:
            fpath = os.path.join(libdir, file)
            if os.path.isfile(fpath):
                good = fpath
                break

        if good is not None:
            resolved.append(good)
        else:
            resolved.append(plugin)

    return paths_sep.join(resolved)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2024 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import threading
import urllib.parse
from enum import Enum
from threading import Lock
from typing import Any, Dict, List, Optional, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from confluent_kafka.schema_registry.schema_registry_client import _RestClient

T = TypeVar("T")


@_attrs_define
class KekKmsProps:
    properties: Dict[str, str] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        field_dict: Dict[str, Any] = {}
        field_dict.update(self.properties)

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        kek_kms_props = cls()

        kek_kms_props.properties = d  # type: ignore[attr-defined]
        return kek_kms_props


@_attrs_define
class Kek:
    name: Optional[str]
    kms_type: Optional[str]
    kms_key_id: Optional[str]
    kms_props: Optional[KekKmsProps]
    doc: Optional[str]
    shared: Optional[bool]
    ts: Optional[int] = _attrs_field(default=None)
    deleted: Optional[bool] = _attrs_field(default=None)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        kms_type = self.kms_type

        kms_key_id = self.kms_key_id

        _kms_props: Optional[Dict[str, Any]] = None
        if self.kms_props is not None:
            _kms_props = self.kms_props.to_dict()

        doc = self.doc

        shared = self.shared

        ts = self.ts

        deleted = self.deleted

        field_dict: Dict[str, Any] = {}
        if name is not None:
            field_dict["name"] = name
        if kms_type is not None:
            field_dict["kmsType"] = kms_type
        if kms_key_id is not None:
            field_dict["kmsKeyId"] = kms_key_id
        if _kms_props is not None:
            field_dict["kmsProps"] = _kms_props
        if doc is not None:
            field_dict["doc"] = doc
        if shared is not None:
            field_dict["shared"] = shared
        if ts is not None:
            field_dict["ts"] = ts
        if deleted is not None:
            field_dict["deleted"] = deleted

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name", None)

        kms_type = d.pop("kmsType", None)

        kms_key_id = d.pop("kmsKeyId", None)

        _kms_props: Optional[Dict[str, Any]] = d.pop("kmsProps", None)
        kms_props: Optional[KekKmsProps]
        if _kms_props is None:
            kms_props = None
        else:
            kms_props = KekKmsProps.from_dict(_kms_props)

        doc = d.pop("doc", None)

        shared = d.pop("shared", None)

        ts = d.pop("ts", None)

        deleted = d.pop("deleted", None)

        kek = cls(  # type: ignore[call-arg]
            name=name,
            kms_type=kms_type,
            kms_key_id=kms_key_id,
            kms_props=kms_props,
            doc=doc,
            shared=shared,
            ts=ts,
            deleted=deleted,
        )

        return kek


@_attrs_define
class CreateKekRequest:
    name: Optional[str]
    kms_type: Optional[str]
    kms_key_id: Optional[str]
    kms_props: Optional[KekKmsProps]
    doc: Optional[str]
    shared: Optional[bool]

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        kms_type = self.kms_type

        kms_key_id = self.kms_key_id

        _kms_props: Optional[Dict[str, Any]] = None
        if self.kms_props is not None:
            _kms_props = self.kms_props.to_dict()

        doc = self.doc

        shared = self.shared

        field_dict: Dict[str, Any] = {}
        if name is not None:
            field_dict["name"] = name
        if kms_type is not None:
            field_dict["kmsType"] = kms_type
        if kms_key_id is not None:
            field_dict["kmsKeyId"] = kms_key_id
        if _kms_props is not None:
            field_dict["kmsProps"] = _kms_props
        if doc is not None:
            field_dict["doc"] = doc
        if shared is not None:
            field_dict["shared"] = shared

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name", None)

        kms_type = d.pop("kmsType", None)

        kms_key_id = d.pop("kmsKeyId", None)

        _kms_props: Optional[Dict[str, Any]] = d.pop("kmsProps", None)
        kms_props: Optional[KekKmsProps]
        if _kms_props is None:
            kms_props = None
        else:
            kms_props = KekKmsProps.from_dict(_kms_props)

        doc = d.pop("doc", None)

        shared = d.pop("shared", None)

        create_kek_request = cls(  # type: ignore[call-arg]
            name=name,
            kms_type=kms_type,
            kms_key_id=kms_key_id,
            kms_props=kms_props,
            doc=doc,
            shared=shared,
        )

        return create_kek_request


class DekAlgorithm(str, Enum):
    AES128_GCM = "AES128_GCM"
    AES256_GCM = "AES256_GCM"
    AES256_SIV = "AES256_SIV"

    def __str__(self) -> str:
        return str(self.value)


@_attrs_define
class Dek:
    kek_name: Optional[str]
    subject: Optional[str]
    version: Optional[int]
    algorithm: Optional[DekAlgorithm]
    encrypted_key_material: Optional[str]
    encrypted_key_material_bytes: Optional[bytes] = _attrs_field(init=False, eq=False, order=False, default=None)
    key_material: Optional[str] = _attrs_field(default=None)
    key_material_bytes: Optional[bytes] = _attrs_field(init=False, eq=False, order=False, default=None)
    ts: Optional[int] = _attrs_field(default=None)
    deleted: Optional[bool] = _attrs_field(default=None)
    _lock: threading.Lock = _attrs_field(factory=threading.Lock, init=False, eq=False, order=False)

    def get_encrypted_key_material_bytes(self) -> Optional[bytes]:
        if self.encrypted_key_material is None:
            return None
        if self.encrypted_key_material_bytes is None:
            with self._lock:
                if self.encrypted_key_material_bytes is None:
                    self.encrypted_key_material_bytes = base64.b64decode(self.encrypted_key_material)
        return self.encrypted_key_material_bytes

    def get_key_material_bytes(self) -> Optional[bytes]:
        if self.key_material is None:
            return None
        if self.key_material_bytes is None:
            with self._lock:
                if self.key_material_bytes is None:
                    self.key_material_bytes = base64.b64decode(self.key_material)
        return self.key_material_bytes

    def set_key_material(self, key_material_bytes: bytes):
        with self._lock:
            if key_material_bytes is None:
                self.key_material = None
            else:
                self.key_material = base64.b64encode(key_material_bytes).decode("utf-8")

    def to_dict(self) -> Dict[str, Any]:
        kek_name = self.kek_name

        subject = self.subject

        version = self.version

        algorithm: Optional[str] = None
        if self.algorithm is not None:
            algorithm = self.algorithm

        encrypted_key_material = self.encrypted_key_material

        key_material = self.key_material

        ts = self.ts

        deleted = self.deleted

        field_dict: Dict[str, Any] = {}
        if kek_name is not None:
            field_dict["kekName"] = kek_name
        if subject is not None:
            field_dict["subject"] = subject
        if version is not None:
            field_dict["version"] = version
        if algorithm is not None:
            field_dict["algorithm"] = algorithm
        if encrypted_key_material is not None:
            field_dict["encryptedKeyMaterial"] = encrypted_key_material
        if key_material is not None:
            field_dict["keyMaterial"] = key_material
        if ts is not None:
            field_dict["ts"] = ts
        if deleted is not None:
            field_dict["deleted"] = deleted

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        kek_name = d.pop("kekName", None)

        subject = d.pop("subject", None)

        version = d.pop("version", None)

        _algorithm = d.pop("algorithm", None)
        algorithm: Optional[DekAlgorithm]
        if _algorithm is None:
            algorithm = None
        else:
            algorithm = DekAlgorithm(_algorithm)

        encrypted_key_material = d.pop("encryptedKeyMaterial", None)

        key_material = d.pop("keyMaterial", None)

        ts = d.pop("ts", None)

        deleted = d.pop("deleted", None)

        dek = cls(  # type: ignore[call-arg]
            kek_name=kek_name,
            subject=subject,
            version=version,
            algorithm=algorithm,
            encrypted_key_material=encrypted_key_material,
            key_material=key_material,
            ts=ts,
            deleted=deleted,
        )

        return dek


@_attrs_define
class CreateDekRequest:
    subject: Optional[str]
    version: Optional[int]
    algorithm: Optional[DekAlgorithm]
    encrypted_key_material: Optional[str]

    def to_dict(self) -> Dict[str, Any]:
        subject = self.subject

        version = self.version

        _algorithm: Optional[str] = None
        if self.algorithm is not None:
            _algorithm = self.algorithm.value

        encrypted_key_material = self.encrypted_key_material

        field_dict: Dict[str, Any] = {}
        if subject is not None:
            field_dict["subject"] = subject
        if version is not None:
            field_dict["version"] = version
        if _algorithm is not None:
            field_dict["algorithm"] = _algorithm
        if encrypted_key_material is not None:
            field_dict["encryptedKeyMaterial"] = encrypted_key_material

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        subject = d.pop("subject", None)

        version = d.pop("version", None)

        _algorithm = d.pop("algorithm", None)
        algorithm: Optional[DekAlgorithm]
        if _algorithm is None:
            algorithm = None
        else:
            algorithm = DekAlgorithm(_algorithm)

        encrypted_key_material = d.pop("encryptedKeyMaterial", None)

        create_dek_request = cls(  # type: ignore[call-arg]
            subject=subject,
            version=version,
            algorithm=algorithm,
            encrypted_key_material=encrypted_key_material,
        )

        return create_dek_request


@_attrs_define(frozen=True)
class KekId:
    name: str
    deleted: bool


@_attrs_define(frozen=True)
class DekId:
    kek_name: str
    subject: str
    version: int
    algorithm: DekAlgorithm
    deleted: bool


class _KekCache(object):
    def __init__(self):
        self.lock = Lock()
        self.keks = {}

    def set(self, kek_id: KekId, kek: Kek):
        with self.lock:
            self.keks[kek_id] = kek

    def get_kek(self, kek_id: KekId) -> Optional[Kek]:
        with self.lock:
            return self.keks.get(kek_id, None)

    def clear(self):
        with self.lock:
            self.keks.clear()


class _DekCache(object):
    def __init__(self):
        self.lock = Lock()
        self.deks = {}

    def set(self, dek_id: DekId, dek: Dek):
        with self.lock:
            self.deks[dek_id] = dek

    def get_dek_ids(self) -> List[DekId]:
        with self.lock:
            return list(self.deks.keys())

    def get_dek(self, dek_id: DekId) -> Optional[Dek]:
        with self.lock:
            return self.deks.get(dek_id, None)

    def remove(self, dek_id: DekId):
        with self.lock:
            if dek_id in self.deks:
                del self.deks[dek_id]

    def clear(self):
        with self.lock:
            self.deks.clear()


class DekRegistryClient(object):
    """
    A Confluent DEK Registry client.

    Args:
        conf (dict): DEK Registry client configuration.

    """  # noqa: E501

    def __init__(self, conf: dict):
        self._conf = conf
        self._rest_client = _RestClient(conf)
        self._kek_cache = _KekCache()
        self._dek_cache = _DekCache()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._rest_client is not None:
            self._rest_client.session.close()

    def config(self):
        return self._conf

    def register_kek(
        self,
        name: str,
        kms_type: str,
        kms_key_id: str,
        shared: bool = False,
        kms_props: Optional[Dict[str, str]] = None,
        doc: Optional[str] = None,
    ) -> Kek:
        """
        Register a new Key Encryption Key (KEK) with the DEK Registry.

        Args:
            name (str): Name of the KEK.
            kms_type (str): Type of the Key Management Service (KMS) used to manage the KEK.
            kms_key_id (str): Identifier of the KEK in the KMS.
            kms_props (Dict[str, str]): Additional properties for the KMS.
            doc (str): Description of the KEK.
            shared (bool): Whether the KEK is shared.

        Returns:
            Kek: KEK instance.

        Raises:
            SchemaRegistryError: If KEK can't be registered.
        """  # noqa: E501

        cache_key = KekId(name=name, deleted=False)
        kek = self._kek_cache.get_kek(cache_key)
        if kek is not None:
            return kek

        request = CreateKekRequest(
            name=name,
            kms_type=kms_type,
            kms_key_id=kms_key_id,
            kms_props=KekKmsProps.from_dict(kms_props) if kms_props is not None else None,
            doc=doc,
            shared=shared,
        )

        response = self._rest_client.post('/dek-registry/v1/keks', request.to_dict())
        kek = Kek.from_dict(response)

        self._kek_cache.set(cache_key, kek)

        return kek

    def get_kek(self, name: str, deleted: bool = False) -> Kek:
        """
        Get a Key Encryption Key (KEK) from the DEK Registry.

        Args:
            name (str): Name of the KEK.
            deleted (bool): Whether to include deleted KEKs.

        Returns:
            Kek: KEK instance.

        Raises:
            SchemaRegistryError: If KEK can't be found.
        """

        cache_key = KekId(name=name, deleted=deleted)
        kek = self._kek_cache.get_kek(cache_key)
        if kek is not None:
            return kek

        query = {'deleted': deleted}
        response = self._rest_client.get('/dek-registry/v1/keks/{}'.format(urllib.parse.quote(name, safe='')), query)
        kek = Kek.from_dict(response)

        self._kek_cache.set(cache_key, kek)

        return kek

    def register_dek(
        self,
        kek_name: str,
        subject: str,
        encrypted_key_material: str,
        algorithm: DekAlgorithm = DekAlgorithm.AES256_GCM,
        version: int = 1,
    ) -> Dek:
        """
        Register a new Data Encryption Key (DEK) with the DEK Registry.

        Args:
            kek_name (str): Name of the Key Encryption Key (KEK) used to encrypt the DEK.
            subject (str): Subject of the DEK.
            version (int): Version of the DEK.
            algorithm (DekAlgorithm): Algorithm used to encrypt the DEK.
            encrypted_key_material (str): Encrypted key material.

        Returns:
            Dek: DEK instance.

        Raises:
            SchemaRegistryError: If DEK can't be registered.
        """

        cache_key = DekId(kek_name=kek_name, subject=subject, version=version, algorithm=algorithm, deleted=False)
        dek = self._dek_cache.get_dek(cache_key)
        if dek is not None:
            return dek

        request = CreateDekRequest(
            subject=subject, version=version, algorithm=algorithm, encrypted_key_material=encrypted_key_material
        )

        dek = self._create_dek(kek_name, request)

        self._dek_cache.set(cache_key, dek)
        # Ensure latest dek is invalidated, such as in case of conflict (409)
        self._dek_cache.remove(
            DekId(kek_name=kek_name, subject=subject, version=-1, algorithm=algorithm, deleted=False)
        )
        self._dek_cache.remove(DekId(kek_name=kek_name, subject=subject, version=-1, algorithm=algorithm, deleted=True))

        return dek

    def _create_dek(self, kek_name: str, request: CreateDekRequest) -> Dek:
        from confluent_kafka.schema_registry.error import SchemaRegistryError

        if request.subject is None:
            raise TypeError("Subject cannot be None")
        try:
            # Try newer API with subject in the path
            path = '/dek-registry/v1/keks/{}/deks/{}'.format(
                urllib.parse.quote(kek_name), urllib.parse.quote(request.subject, safe='')
            )
            response = self._rest_client.post(path, request.to_dict())
            return Dek.from_dict(response)
        except SchemaRegistryError as e:
            if e.http_status_code == 405:
                # Try fallback to older API that does not have subject in the path
                path = '/dek-registry/v1/keks/{}/deks'.format(urllib.parse.quote(kek_name))
                response = self._rest_client.post(path, request.to_dict())
                return Dek.from_dict(response)
            else:
                raise

    def get_dek(
        self,
        kek_name: str,
        subject: str,
        algorithm: DekAlgorithm = DekAlgorithm.AES256_GCM,
        version: int = 1,
        deleted: bool = False,
    ) -> Dek:
        """
        Get a Data Encryption Key (DEK) from the DEK Registry.

        Args:
            kek_name (str): Name of the Key Encryption Key (KEK) used to encrypt the DEK.
            subject (str): Subject of the DEK.
            version (int): Version of the DEK.
            algorithm (DekAlgorithm): Algorithm used to encrypt the DEK.
            deleted (bool): Whether to include deleted DEKs.

        Returns:
            Dek: DEK instance.

        Raises:
            SchemaRegistryError: If DEK can't be found.
        """

        cache_key = DekId(kek_name=kek_name, subject=subject, version=version, algorithm=algorithm, deleted=deleted)
        dek = self._dek_cache.get_dek(cache_key)
        if dek is not None:
            return dek

        query = {'algorithm': algorithm, 'deleted': deleted}
        response = self._rest_client.get(
            '/dek-registry/v1/keks/{}/deks/{}/versions/{}'.format(
                urllib.parse.quote(kek_name), urllib.parse.quote(subject, safe=''), version
            ),
            query,
        )
        dek = Dek.from_dict(response)

        self._dek_cache.set(cache_key, dek)

        return dek

    @staticmethod
    def new_client(conf: dict) -> 'DekRegistryClient':
        from .mock_dek_registry_client import MockDekRegistryClient

        url = conf.get("url")
        if url is None:
            return MockDekRegistryClient({"url": "mock://"})
        if url.startswith("mock://"):
            return MockDekRegistryClient(conf)
        return DekRegistryClient(conf)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import json
from typing import Any, Callable, Coroutine, Dict, Optional, Union, cast

from fastavro import schemaless_reader, schemaless_writer

from confluent_kafka.schema_registry import (
    AsyncSchemaRegistryClient,
    RuleMode,
    Schema,
    dual_schema_id_deserializer,
    prefix_schema_id_serializer,
    topic_subject_name_strategy,
)
from confluent_kafka.schema_registry.common import asyncinit
from confluent_kafka.schema_registry.common.avro import (
    AVRO_TYPE,
    AvroSchema,
    _ContextStringIO,
    _schema_loads,
    get_inline_tags,
    parse_schema_with_repo,
    transform,
)
from confluent_kafka.schema_registry.common.schema_registry_client import RulePhase
from confluent_kafka.schema_registry.rule_registry import RuleRegistry
from confluent_kafka.schema_registry.serde import (
    AsyncBaseDeserializer,
    AsyncBaseSerializer,
    ParsedSchemaCache,
    SchemaId,
)
from confluent_kafka.serialization import SerializationContext, SerializationError

__all__ = [
    '_resolve_named_schema',
    'AsyncAvroSerializer',
    'AsyncAvroDeserializer',
]


async def _resolve_named_schema(
    schema: Schema, schema_registry_client: AsyncSchemaRegistryClient
) -> Dict[str, AvroSchema]:
    """
    Resolves named schemas referenced by the provided schema recursively.
    :param schema: Schema to resolve named schemas for.
    :param schema_registry_client: SchemaRegistryClient to use for retrieval.
    :return: named_schemas dict.
    """
    named_schemas = {}
    if schema.references is not None:
        for ref in schema.references:
            if ref.subject is None or ref.version is None:
                raise TypeError("Subject or version cannot be None")
            referenced_schema = await schema_registry_client.get_version(ref.subject, ref.version, True)
            ref_named_schemas = await _resolve_named_schema(referenced_schema.schema, schema_registry_client)
            if referenced_schema.schema.schema_str is None:
                raise TypeError("Schema string cannot be None")

            parsed_schema = parse_schema_with_repo(referenced_schema.schema.schema_str, named_schemas=ref_named_schemas)
            named_schemas.update(ref_named_schemas)
            if ref.name is None:
                raise TypeError("Name cannot be None")
            named_schemas[ref.name] = parsed_schema
    return named_schemas


@asyncinit
class AsyncAvroSerializer(AsyncBaseSerializer):
    """
    Serializer that outputs Avro binary encoded data with Confluent Schema Registry framing.

    Configuration properties:

    +-----------------------------------+----------+--------------------------------------------------+
    | Property Name                     | Type     | Description                                      |
    +===================================+==========+==================================================+
    | ``auto.register.schemas``         | bool     | If True, automatically register the configured   |
    |                                   |          | schema with Confluent Schema Registry if it has  |
    |                                   |          | not previously been associated with the relevant |
    |                                   |          | subject (determined via subject.name.strategy).  |
    |                                   |          |                                                  |
    |                                   |          | Defaults to True.                                |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``normalize.schemas``             | bool     | Whether to normalize schemas, which will         |
    |                                   |          | transform schemas to have a consistent format,   |
    |                                   |          | including ordering properties and references.    |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``use.schema.id``                 | int      | Whether to use the given schema ID for           |
    |                                   |          | serialization.                                   |
    |                                   |          |                                                  |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``use.latest.version``            | bool     | Whether to use the latest subject version for    |
    |                                   |          | serialization.                                   |
    |                                   |          |                                                  |
    |                                   |          | WARNING: There is no check that the latest       |
    |                                   |          | schema is backwards compatible with the object   |
    |                                   |          | being serialized.                                |
    |                                   |          |                                                  |
    |                                   |          | Defaults to False.                               |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``use.latest.with.metadata``      | dict     | Whether to use the latest subject version with   |
    |                                   |          | the given metadata.                              |
    |                                   |          |                                                  |
    |                                   |          | WARNING: There is no check that the latest       |
    |                                   |          | schema is backwards compatible with the object   |
    |                                   |          | being serialized.                                |
    |                                   |          |                                                  |
    |                                   |          | Defaults to None.                                |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``subject.name.strategy``         | callable | Callable(SerializationContext, str) -> str       |
    |                                   |          |                                                  |
    |                                   |          | Defines how Schema Registry subject names are    |
    |                                   |          | constructed. Standard naming strategies are      |
    |                                   |          | defined in the confluent_kafka.schema_registry   |
    |                                   |          | namespace.                                       |
    |                                   |          |                                                  |
    |                                   |          | Defaults to topic_subject_name_strategy.         |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``schema.id.serializer``          | callable | Callable(bytes, SerializationContext, schema_id) |
    |                                   |          |   -> bytes                                       |
    |                                   |          |                                                  |
    |                                   |          | Defines how the schema id/guid is serialized.    |
    |                                   |          | Defaults to prefix_schema_id_serializer.         |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``validate.strict``               | bool     | If set to True, an error will be raised if       |
    |                                   |          | records do not contain exactly the same          |
    |                                   |          | fields that the schema states.                   |
    |                                   |          |                                                  |
    |                                   |          | Defaults to False.                               |
    +-----------------------------------+----------+--------------------------------------------------+
    | ``validate.strict.allow.default`` | bool     | If set to True, an error will be raised          |
    |                                   |          | if records do not contain exactly the same       |
    |                                   |          | fields that the schema states, unless it is a    |
    |                                   |          | missing field that has a default value in the    |
    |                                   |          | schema.                                          |
    |                                   |          |                                                  |
    |                                   |          | Defaults to False.                               |
    +-----------------------------------+----------+--------------------------------------------------+

    Schemas are registered against subject names in Confluent Schema Registry that
    define a scope in which the schemas can be evolved. By default, the subject name
    is formed by concatenating the topic name with the message field (key or value)
    separated by a hyphen.

    i.e. {topic name}-{message field}

    Alternative naming strategies may be configured with the property
    ``subject.name.strategy``.

    Supported subject name strategies:

    +--------------------------------------+------------------------------+
    | Subject Name Strategy                | Output Format                |
    +======================================+==============================+
    | topic_subject_name_strategy(default) | {topic name}-{message field} |
    +--------------------------------------+------------------------------+
    | topic_record_subject_name_strategy   | {topic name}-{record name}   |
    +--------------------------------------+------------------------------+
    | record_subject_name_strategy         | {record name}                |
    +--------------------------------------+------------------------------+

    See `Subject name strategy <https://docs.confluent.io/current/schema-registry/serializer-formatter.html#subject-name-strategy>`_ for additional details.

    Note:
        Prior to serialization, all values must first be converted to
        a dict instance. This may handled manually prior to calling
        :py:func:`Producer.produce()` or by registering a `to_dict`
        callable with AvroSerializer.

        See ``avro_producer.py`` in the examples directory for example usage.

    Note:
       Tuple notation can be used to determine which branch of an ambiguous union to take.

       See `fastavro notation <https://fastavro.readthedocs.io/en/latest/writer.html#using-the-tuple-notation-to-specify-which-branch-of-a-union-to-take>`_

    Args:
        schema_registry_client (SchemaRegistryClient): Schema Registry client instance.

        schema_str (str or Schema):
            Avro `Schema Declaration. <https://avro.apache.org/docs/current/spec.html#schemas>`_
            Accepts either a string or a :py:class:`Schema` instance. Note that string
            definitions cannot reference other schemas. For referencing other schemas,
            use a :py:class:`Schema` instance.

        to_dict (callable, optional): Callable(object, SerializationContext) -> dict. Converts object to a dict.

        conf (dict): AvroSerializer configuration.
    """  # noqa: E501

    __slots__ = [
        '_known_subjects',
        '_parsed_schema',
        '_schema',
        '_schema_id',
        '_schema_name',
        '_to_dict',
        '_parsed_schemas',
        '_strict',
        '_strict_allow_default',
    ]

    _default_conf = {
        'auto.register.schemas': True,
        'normalize.schemas': False,
        'use.schema.id': None,
        'use.latest.version': False,
        'use.latest.with.metadata': None,
        'subject.name.strategy': topic_subject_name_strategy,
        'schema.id.serializer': prefix_schema_id_serializer,
        'validate.strict': False,
        'validate.strict.allow.default': False,
    }

    async def __init_impl(
        self,
        schema_registry_client: AsyncSchemaRegistryClient,
        schema_str: Union[str, Schema, None] = None,
        to_dict: Optional[Callable[[object, SerializationContext], dict]] = None,
        conf: Optional[dict] = None,
        rule_conf: Optional[dict] = None,
        rule_registry: Optional[RuleRegistry] = None,
    ):
        super().__init__()
        if isinstance(schema_str, str):
            schema = _schema_loads(schema_str)
        elif isinstance(schema_str, Schema):
            schema = schema_str
        else:
            schema = None

        self._registry = schema_registry_client
        self._schema_id: Optional[SchemaId] = None
        self._rule_registry = rule_registry if rule_registry else RuleRegistry.get_global_instance()
        self._known_subjects: set[str] = set()
        self._parsed_schemas = ParsedSchemaCache()

        if to_dict is not None and not callable(to_dict):
            raise ValueError(
                "to_dict must be callable with the signature " "to_dict(object, SerializationContext)->dict"
            )

        self._to_dict = to_dict

        conf_copy = self._default_conf.copy()
        if conf is not None:
            conf_copy.update(conf)

        self._auto_register = cast(bool, conf_copy.pop('auto.register.schemas'))
        if not isinstance(self._auto_register, bool):
            raise ValueError("auto.register.schemas must be a boolean value")

        self._normalize_schemas = cast(bool, conf_copy.pop('normalize.schemas'))
        if not isinstance(self._normalize_schemas, bool):
            raise ValueError("normalize.schemas must be a boolean value")

        self._use_schema_id = cast(Optional[int], conf_copy.pop('use.schema.id'))
        if self._use_schema_id is not None and not isinstance(self._use_schema_id, int):
            raise ValueError("use.schema.id must be an int value")

        self._use_latest_version = cast(bool, conf_copy.pop('use.latest.version'))
        if not isinstance(self._use_latest_version, bool):
            raise ValueError("use.latest.version must be a boolean value")
        if self._use_latest_version and self._auto_register:
            raise ValueError("cannot enable both use.latest.version and auto.register.schemas")

        self._use_latest_with_metadata = cast(Optional[dict], conf_copy.pop('use.latest.with.metadata'))
        if self._use_latest_with_metadata is not None and not isinstance(self._use_latest_with_metadata, dict):
            raise ValueError("use.latest.with.metadata must be a dict value")

        self._subject_name_func = cast(
            Callable[[Optional[SerializationContext], Optional[str]], Optional[str]],
            conf_copy.pop('subject.name.strategy'),
        )
        if not callable(self._subject_name_func):
            raise ValueError("subject.name.strategy must be callable")

        self._schema_id_serializer = cast(
            Callable[[bytes, Optional[SerializationContext], Any], bytes], conf_copy.pop('schema.id.serializer')
        )
        if not callable(self._schema_id_serializer):
            raise ValueError("schema.id.serializer must be callable")

        self._strict = cast(bool, conf_copy.pop('validate.strict'))
        if not isinstance(self._strict, bool):
            raise ValueError("validate.strict must be a boolean value")

        self._strict_allow_default = cast(bool, conf_copy.pop('validate.strict.allow.default'))
        if not isinstance(self._strict_allow_default, bool):
            raise ValueError("validate.strict.allow.default must be a boolean value")

        if len(conf_copy) > 0:
            raise ValueError("Unrecognized properties: {}".format(", ".join(conf_copy.keys())))

        if schema:
            parsed_schema = await self._get_parsed_schema(schema)

            if isinstance(parsed_schema, list):
                # if parsed_schema is a list, we have an Avro union and there
                # is no valid schema name. This is fine because the only use of
                # schema_name is for supplying the subject name to the registry
                # and union types should use topic_subject_name_strategy, which
                # just discards the schema name anyway
                schema_name = None
            elif isinstance(parsed_schema, dict):
                # The Avro spec states primitives have a name equal to their type
                # i.e. {"type": "string"} has a name of string.
                # This function does not comply.
                # https://github.com/fastavro/fastavro/issues/415
                if schema.schema_str is not None:
                    schema_dict = json.loads(schema.schema_str)
                    schema_name = parsed_schema.get("name", schema_dict.get("type"))
                else:
                    schema_name = None
            else:
                schema_name = None
        else:
            schema_name = None
            parsed_schema = None

        self._schema = schema
        self._schema_name = schema_name
        self._parsed_schema = parsed_schema

        for rule in self._rule_registry.get_executors():
            rule.configure(self._registry.config() if self._registry else {}, rule_conf if rule_conf else {})

    __init__ = __init_impl

    def __call__(  # type: ignore[override]
        self, obj: object, ctx: Optional[SerializationContext] = None
    ) -> Coroutine[Any, Any, Optional[bytes]]:
        return self.__serialize(obj, ctx)

    async def __serialize(self, obj: object, ctx: Optional[SerializationContext] = None) -> Optional[bytes]:
        """
        Serializes an object to Avro binary format, prepending it with Confluent
        Schema Registry framing.

        Args:
            obj (object): The object instance to serialize.

            ctx (SerializationContext): Metadata pertaining to the serialization operation.

        Raises:
            TypeError or ValueError: If any error occurs serializing obj.
            SchemaRegistryError: If there was an error registering the schema with
                                 Schema Registry, or auto.register.schemas is
                                 false and the schema was not registered.

        Returns:
            bytes: Confluent Schema Registry encoded Avro bytes
        """

        if obj is None:
            return None

        subject = self._subject_name_func(ctx, self._schema_name)
        latest_schema = await self._get_reader_schema(subject) if subject else None
        if latest_schema is not None:
            self._schema_id = SchemaId(AVRO_TYPE, latest_schema.schema_id, latest_schema.guid)
        elif subject is not None and subject not in self._known_subjects:
            # Check to ensure this schema has been registered under subject_name.
            if self._auto_register:
                # The schema name will always be the same. We can't however register
                # a schema without a subject so we set the schema_id here to handle
                # the initial registration.
                registered_schema = await self._registry.register_schema_full_response(
                    subject, self._schema, normalize_schemas=self._normalize_schemas
                )
                self._schema_id = SchemaId(AVRO_TYPE, registered_schema.schema_id, registered_schema.guid)
            else:
                registered_schema = await self._registry.lookup_schema(
                    subject, self._schema, normalize_schemas=self._normalize_schemas
                )
                self._schema_id = SchemaId(AVRO_TYPE, registered_schema.schema_id, registered_schema.guid)

            self._known_subjects.add(subject)

        value: Any
        parsed_schema: Any
        if self._to_dict is not None:
            if ctx is None:
                raise TypeError("SerializationContext cannot be None")
            value = self._to_dict(obj, ctx)
        else:
            value = obj

        if latest_schema is not None and ctx is not None and subject is not None:
            parsed_schema = await self._get_parsed_schema(latest_schema.schema)

            def field_transformer(rule_ctx, field_transform, msg):
                return transform(rule_ctx, parsed_schema, msg, field_transform)  # noqa: E731

            value = self._execute_rules(
                ctx,
                subject,
                RuleMode.WRITE,
                None,
                latest_schema.schema,
                value,
                get_inline_tags(parsed_schema),
                field_transformer,
            )
        else:
            parsed_schema = self._parsed_schema

        with _ContextStringIO() as fo:
            # Check if it's a simple bytes type
            is_bytes = parsed_schema == "bytes" or (
                isinstance(parsed_schema, dict) and parsed_schema.get("type") == "bytes"
            )
            if is_bytes:
                # For simple bytes type, write value directly
                buffer = value if isinstance(value, bytes) else value.encode()
            else:
                # write the record to the rest of the buffer
                schemaless_writer(
                    fo, parsed_schema, value, strict=self._strict, strict_allow_default=self._strict_allow_default
                )
                buffer = fo.getvalue()

            if latest_schema is not None and ctx is not None and subject is not None:
                buffer = self._execute_rules_with_phase(
                    ctx, subject, RulePhase.ENCODING, RuleMode.WRITE, None, latest_schema.schema, buffer, None, None
                )

            return self._schema_id_serializer(buffer, ctx, self._schema_id)

    async def _get_parsed_schema(self, schema: Schema) -> AvroSchema:
        parsed_schema = self._parsed_schemas.get_parsed_schema(schema)
        if parsed_schema is not None:
            return parsed_schema

        named_schemas = await _resolve_named_schema(schema, self._registry)
        if schema.schema_str is None:
            raise TypeError("Schema string cannot be None")
        prepared_schema = _schema_loads(schema.schema_str)
        if prepared_schema.schema_str is None:
            raise TypeError("Prepared schema string cannot be None")
        parsed_schema = parse_schema_with_repo(prepared_schema.schema_str, named_schemas=named_schemas)

        self._parsed_schemas.set(schema, parsed_schema)
        return parsed_schema


@asyncinit
class AsyncAvroDeserializer(AsyncBaseDeserializer):
    """
    Deserializer for Avro binary encoded data with Confluent Schema Registry
    framing.

    +-----------------------------+----------+--------------------------------------------------+
    | Property Name               | Type     | Description                                      |
    +-----------------------------+----------+--------------------------------------------------+
    |                             |          | Whether to use the latest subject version for    |
    | ``use.latest.version``      | bool     | deserialization.                                 |
    |                             |          |                                                  |
    |                             |          | Defaults to False.                               |
    +-----------------------------+----------+--------------------------------------------------+
    |                             |          | Whether to use the latest subject version with   |
    | ``use.latest.with.metadata``| dict     | the given metadata.                              |
    |                             |          |                                                  |
    |                             |          | Defaults to None.                                |
    +-----------------------------+----------+--------------------------------------------------+
    |                             |          | Callable(SerializationContext, str) -> str       |
    |                             |          |                                                  |
    | ``subject.name.strategy``   | callable | Defines how Schema Registry subject names are    |
    |                             |          | constructed. Standard naming strategies are      |
    |                             |          | defined in the confluent_kafka.schema_registry   |
    |                             |          | namespace.                                       |
    |                             |          |                                                  |
    |                             |          | Defaults to topic_subject_name_strategy.         |
    +-----------------------------+----------+--------------------------------------------------+
    |                             |          | Callable(bytes, SerializationContext, schema_id) |
    |                             |          |   -> io.BytesIO                                  |
    |                             |          |                                                  |
    | ``schema.id.deserializer``  | callable | Defines how the schema id/guid is deserialized.  |
    |                             |          | Defaults to dual_schema_id_deserializer.         |
    +-----------------------------+----------+--------------------------------------------------+

    Note:
        By default, Avro complex types are returned as dicts. This behavior can
        be overridden by registering a callable ``from_dict`` with the deserializer to
        convert the dicts to the desired type.

        See ``avro_consumer.py`` in the examples directory in the examples
        directory for example usage.

    Args:
        schema_registry_client (SchemaRegistryClient): Confluent Schema Registry
            client instance.

        schema_str (str, Schema, optional): Avro reader schema declaration Accepts
            either a string or a :py:class:`Schema` instance. If not provided, the
            writer schema will be used as the reader schema. Note that string
            definitions cannot reference other schemas. For referencing other schemas,
            use a :py:class:`Schema` instance.

        from_dict (callable, optional): Callable(dict, SerializationContext) -> object.
            Converts a dict to an instance of some object.

        return_record_name (bool): If True, when reading a union of records, the result will
                                   be a tuple where the first value is the name of the record and the second value is
                                   the record itself.  Defaults to False.

    See Also:
        `Apache Avro Schema Declaration <https://avro.apache.org/docs/current/spec.html#schemas>`_

        `Apache Avro Schema Resolution <https://avro.apache.org/docs/1.8.2/spec.html#Schema+Resolution>`_
    """

    __slots__ = ['_reader_schema', '_from_dict', '_return_record_name', '_schema', '_parsed_schemas']

    _default_conf = {
        'use.latest.version': False,
        'use.latest.with.metadata': None,
        'subject.name.strategy': topic_subject_name_strategy,
        'schema.id.deserializer': dual_schema_id_deserializer,
    }

    async def __init_impl(
        self,
        schema_registry_client: AsyncSchemaRegistryClient,
        schema_str: Union[str, Schema, None] = None,
        from_dict: Optional[Callable[[dict, SerializationContext], object]] = None,
        return_record_name: bool = False,
        conf: Optional[dict] = None,
        rule_conf: Optional[dict] = None,
        rule_registry: Optional[RuleRegistry] = None,
    ):
        super().__init__()
        schema = None
        if schema_str is not None:
            if isinstance(schema_str, str):
                schema = _schema_loads(schema_str)
            elif isinstance(schema_str, Schema):
                schema = schema_str
            else:
                raise TypeError('You must pass either schema string or schema object')

        self._schema = schema
        self._registry = schema_registry_client
        self._rule_registry = rule_registry if rule_registry else RuleRegistry.get_global_instance()
        self._parsed_schemas = ParsedSchemaCache()
        self._use_schema_id = None

        conf_copy = self._default_conf.copy()
        if conf is not None:
            conf_copy.update(conf)

        self._use_latest_version = cast(bool, conf_copy.pop('use.latest.version'))
        if not isinstance(self._use_latest_version, bool):
            raise ValueError("use.latest.version must be a boolean value")

        self._use_latest_with_metadata = cast(Optional[dict], conf_copy.pop('use.latest.with.metadata'))
        if self._use_latest_with_metadata is not None and not isinstance(self._use_latest_with_metadata, dict):
            raise ValueError("use.latest.with.metadata must be a dict value")

        self._subject_name_func = cast(
            Callable[[Optional[SerializationContext], Optional[str]], Optional[str]],
            conf_copy.pop('subject.name.strategy'),
        )
        if not callable(self._subject_name_func):
            raise ValueError("subject.name.strategy must be callable")

        self._schema_id_deserializer = cast(
            Callable[[bytes, Optional[SerializationContext], Any], io.BytesIO], conf_copy.pop('schema.id.deserializer')
        )
        if not callable(self._schema_id_deserializer):
            raise ValueError("schema.id.deserializer must be callable")

        if len(conf_copy) > 0:
            raise ValueError("Unrecognized properties: {}".format(", ".join(conf_copy.keys())))

        self._reader_schema: Optional[AvroSchema]
        if schema and self._schema is not None:
            self._reader_schema = await self._get_parsed_schema(self._schema)
        else:
            self._reader_schema = None

        if from_dict is not None and not callable(from_dict):
            raise ValueError(
                "from_dict must be callable with the signature " "from_dict(SerializationContext, dict) -> object"
            )
        self._from_dict = from_dict

        self._return_record_name = return_record_name
        if not isinstance(self._return_record_name, bool):
            raise ValueError("return_record_name must be a boolean value")

        for rule in self._rule_registry.get_executors():
            rule.configure(self._registry.config() if self._registry else {}, rule_conf if rule_conf else {})

    __init__ = __init_impl

    def __call__(
        self, data: Optional[bytes], ctx: Optional[SerializationContext] = None
    ) -> Coroutine[Any, Any, Union[dict, object, None]]:
        return self.__deserialize(data, ctx)

    async def __deserialize(
        self, data: Optional[bytes], ctx: Optional[SerializationContext] = None
    ) -> Union[dict, object, None]:
        """
        Deserialize Avro binary encoded data with Confluent Schema Registry framing to
        a dict, or object instance according to from_dict, if specified.

        Arguments:
            data (bytes): bytes

            ctx (SerializationContext): Metadata relevant to the serialization
                operation.

        Raises:
            SerializerError: if an error occurs parsing data.

        Returns:
            object: If data is None, then None. Else, a dict, or object instance according
                    to from_dict, if specified.
        """  # noqa: E501

        if data is None:
            return None

        if len(data) <= 5:
            raise SerializationError(
                "Expecting data framing of length 6 bytes or "
                "more but total data size is {} bytes. This "
                "message was not produced with a Confluent "
                "Schema Registry serializer".format(len(data))
            )

        subject = self._subject_name_func(ctx, None) if ctx else None
        latest_schema = None
        if subject is not None:
            latest_schema = await self._get_reader_schema(subject)

        schema_id = SchemaId(AVRO_TYPE)
        payload = self._schema_id_deserializer(data, ctx, schema_id)

        writer_schema_raw = await self._get_writer_schema(schema_id, subject)
        writer_schema = await self._get_parsed_schema(writer_schema_raw)
        if subject is None:
            subject = (
                self._subject_name_func(ctx, writer_schema.get("name")) if ctx else None  # type: ignore[union-attr]
            )
            if subject is not None:
                latest_schema = await self._get_reader_schema(subject)

        if ctx is not None and subject is not None:
            payload = self._execute_rules_with_phase(
                ctx, subject, RulePhase.ENCODING, RuleMode.READ, None, writer_schema_raw, payload, None, None
            )
        if isinstance(payload, bytes):
            payload = io.BytesIO(payload)

        reader_schema: Optional[AvroSchema]
        if latest_schema is not None and subject is not None:
            migrations = await self._get_migrations(subject, writer_schema_raw, latest_schema, None)
            reader_schema_raw = latest_schema.schema
            reader_schema = await self._get_parsed_schema(latest_schema.schema)
        elif self._schema is not None:
            migrations = None
            reader_schema_raw = self._schema
            reader_schema = self._reader_schema
        else:
            migrations = None
            reader_schema_raw = writer_schema_raw
            reader_schema = writer_schema

        # Check if it's a simple bytes type
        is_bytes = writer_schema == "bytes" or (
            isinstance(writer_schema, dict) and writer_schema.get("type") == "bytes"
        )

        if migrations and ctx is not None and subject is not None:
            if is_bytes:
                # For simple bytes type, read payload directly
                obj_dict = payload.read()
            else:
                obj_dict = schemaless_reader(payload, writer_schema, None, self._return_record_name)
            obj_dict = self._execute_migrations(ctx, subject, migrations, obj_dict)
        else:
            if is_bytes:
                # For simple bytes type, read payload directly
                obj_dict = payload.read()
            else:
                obj_dict = schemaless_reader(payload, writer_schema, reader_schema, self._return_record_name)

        def field_transformer(rule_ctx, field_transform, message):
            return transform(rule_ctx, reader_schema, message, field_transform)  # noqa: E731

        if ctx is not None and subject is not None:
            inline_tags = get_inline_tags(reader_schema) if reader_schema is not None else None
            obj_dict = self._execute_rules(
                ctx, subject, RuleMode.READ, None, reader_schema_raw, obj_dict, inline_tags, field_transformer
            )

        if self._from_dict is not None:
            if ctx is None:
                raise TypeError("SerializationContext cannot be None")
            return self._from_dict(obj_dict, ctx)

        return obj_dict

    async def _get_parsed_schema(self, schema: Schema) -> AvroSchema:
        parsed_schema = self._parsed_schemas.get_parsed_schema(schema)
        if parsed_schema is not None:
            return parsed_schema

        named_schemas = await _resolve_named_schema(schema, self._registry)
        if schema.schema_str is None:
            raise TypeError("Schema string cannot be None")
        prepared_schema = _schema_loads(schema.schema_str)
        if prepared_schema.schema_str is None:
            raise TypeError("Prepared schema string cannot be None")
        parsed_schema = parse_schema_with_repo(prepared_schema.schema_str, named_schemas=named_schemas)

        self._parsed_schemas.set(schema, parsed_schema)
        return parsed_schema



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Dict, Optional

from confluent_kafka.cimpl import Producer as _ProducerImpl

from ._types import DeliveryCallback, HeadersType
from .error import KeySerializationError, ValueSerializationError
from .serialization import MessageField, SerializationContext


class SerializingProducer(_ProducerImpl):
    """
    A high level Kafka producer with serialization capabilities.

    `This class is experimental and likely to be removed, or subject to incompatible API
    changes in future versions of the library. To avoid breaking changes on upgrading, we
    recommend using serializers directly.`

    Derived from the :py:class:`Producer` class, overriding the :py:func:`Producer.produce`
    method to add serialization capabilities.

    Additional configuration properties:

    +-------------------------+---------------------+-----------------------------------------------------+
    | Property Name           | Type                | Description                                         |
    +=========================+=====================+=====================================================+
    |                         |                     | Callable(obj, SerializationContext) -> bytes        |
    | ``key.serializer``      | callable            |                                                     |
    |                         |                     | Serializer used for message keys.                   |
    +-------------------------+---------------------+-----------------------------------------------------+
    |                         |                     | Callable(obj, SerializationContext) -> bytes        |
    | ``value.serializer``    | callable            |                                                     |
    |                         |                     | Serializer used for message values.                 |
    +-------------------------+---------------------+-----------------------------------------------------+

    Serializers for string, integer and double (:py:class:`StringSerializer`, :py:class:`IntegerSerializer`
    and :py:class:`DoubleSerializer`) are supplied out-of-the-box in the ``confluent_kafka.serialization``
    namespace.

    Serializers for Protobuf, JSON Schema and Avro (:py:class:`ProtobufSerializer`, :py:class:`JSONSerializer`
    and :py:class:`AvroSerializer`) with Confluent Schema Registry integration are supplied out-of-the-box
    in the ``confluent_kafka.schema_registry`` namespace.

    See Also:
        - The :ref:`Configuration Guide <pythonclient_configuration>` for in depth information on how to configure the client.
        - `CONFIGURATION.md <https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md>`_ for a comprehensive set of configuration properties.
        - `STATISTICS.md <https://github.com/edenhill/librdkafka/blob/master/STATISTICS.md>`_ for detailed information on the statistics provided by stats_cb
        - The :py:class:`Producer` class for inherited methods.

    Args:
        conf (producer): SerializingProducer configuration.
    """  # noqa E501

    def __init__(self, conf: Dict[str, Any]) -> None:
        conf_copy = conf.copy()

        self._key_serializer = conf_copy.pop('key.serializer', None)
        self._value_serializer = conf_copy.pop('value.serializer', None)

        super(SerializingProducer, self).__init__(conf_copy)

    def produce(  # type: ignore[override]
        self,
        topic: str,
        key: Any = None,
        value: Any = None,
        partition: int = -1,
        on_delivery: Optional[DeliveryCallback] = None,
        timestamp: int = 0,
        headers: Optional[HeadersType] = None,
    ) -> None:
        """
        Produce a message.

        This is an asynchronous operation. An application may use the
        ``on_delivery`` argument to pass a function (or lambda) that will be
        called from :py:func:`SerializingProducer.poll` when the message has
        been successfully delivered or permanently fails delivery.

        Note:
            Currently message headers are not supported on the message returned to
            the callback. The ``msg.headers()`` will return None even if the
            original message had headers set.

        Args:
            topic (str): Topic to produce message to.

            key (object, optional): Message payload key.

            value (object, optional): Message payload value.

            partition (int, optional): Partition to produce to, else the
                configured built-in partitioner will be used.

            on_delivery (callable(KafkaError, Message), optional): Delivery
                report callback. Called as a side effect of
                :py:func:`SerializingProducer.poll` or
                :py:func:`SerializingProducer.flush` on successful or
                failed delivery.

            timestamp (int, optional): Message timestamp (CreateTime) in
                milliseconds since Unix epoch UTC (requires broker >= 0.10.0.0).
                Default value is current time.

            headers (dict, optional): Message headers. The header key must be
                a str while the value must be binary, unicode or None. (Requires
                broker version >= 0.11.0.0)

        Raises:
            BufferError: if the internal producer message queue is full.
                (``queue.buffering.max.messages`` exceeded). If this happens
                the application should call :py:func:`SerializingProducer.Poll`
                and try again.

            KeySerializationError: If an error occurs during key serialization.

            ValueSerializationError: If an error occurs during value serialization.

            KafkaException: For all other errors
        """

        ctx = SerializationContext(topic, MessageField.KEY, headers)
        if self._key_serializer is not None:
            try:
                key = self._key_serializer(key, ctx)
            except Exception as se:
                raise KeySerializationError(se)
        ctx.field = MessageField.VALUE
        if self._value_serializer is not None:
            try:
                value = self._value_serializer(value, ctx)
            except Exception as se:
                raise ValueSerializationError(se)

        super(SerializingProducer, self).produce(
            topic, value, key, headers=headers, partition=partition, timestamp=timestamp, on_delivery=on_delivery
        )



# Copyright 2024 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import io
import logging
import time
from typing import Any, List, Optional, Tuple

from tink import KmsClient, TinkError, aead, daead, kms_client_from_uri, register_kms_client
from tink.core import Registry
from tink.proto import aes_siv_pb2, tink_pb2

from confluent_kafka.schema_registry import _MAGIC_BYTE_V0, RuleMode, SchemaRegistryError
from confluent_kafka.schema_registry.rule_registry import RuleRegistry
from confluent_kafka.schema_registry.rules.encryption.dek_registry.dek_registry_client import (
    Dek,
    DekAlgorithm,
    DekId,
    DekRegistryClient,
    Kek,
    KekId,
)
from confluent_kafka.schema_registry.rules.encryption.kms_driver_registry import KmsDriver, get_kms_driver
from confluent_kafka.schema_registry.serde import (
    FieldContext,
    FieldRuleExecutor,
    FieldTransform,
    FieldType,
    RuleContext,
    RuleError,
    RuleExecutor,
)

log = logging.getLogger(__name__)


aead.register()
daead.register()

ENCRYPT_KEK_NAME = "encrypt.kek.name"
ENCRYPT_KMS_KEY_ID = "encrypt.kms.key.id"
ENCRYPT_KMS_TYPE = "encrypt.kms.type"
ENCRYPT_DEK_ALGORITHM = "encrypt.dek.algorithm"
ENCRYPT_DEK_EXPIRY_DAYS = "encrypt.dek.expiry.days"
ENCRYPT_ALTERNATE_KMS_KEY_IDS = "encrypt.alternate.kms.key.ids"

MILLIS_IN_DAY = 24 * 60 * 60 * 1000


class Clock(object):
    def now(self) -> int:
        return int(round(time.time() * 1000))


class EncryptionExecutor(RuleExecutor):

    def __init__(self, clock: Clock = Clock()):
        self.client: Optional[DekRegistryClient] = None
        self.config: Optional[dict] = None
        self.clock = clock

    def configure(self, client_conf: dict, rule_conf: dict):
        if client_conf:
            if self.client:
                if self.client.config() != client_conf:
                    raise RuleError("executor already configured")
            else:
                self.client = DekRegistryClient.new_client(client_conf)

        if self.config:
            if rule_conf:
                for key, value in rule_conf.items():
                    v = self.config.get(key)
                    if v is not None:
                        if v != value:
                            raise RuleError(f"rule config key already set: {key}")
                    else:
                        self.config[key] = value
        else:
            self.config = rule_conf if rule_conf else {}

    def type(self) -> str:
        return "ENCRYPT_PAYLOAD"

    def transform(self, ctx: RuleContext, message: Any) -> Any:
        executor = self.new_transform(ctx)
        return executor.transform(ctx, FieldType.BYTES, message)

    def new_transform(self, ctx: RuleContext) -> 'EncryptionExecutorTransform':
        cryptor = self._get_cryptor(ctx)
        kek_name = self._get_kek_name(ctx)
        dek_expiry_days = self._get_dek_expiry_days(ctx)
        transform = EncryptionExecutorTransform(self, cryptor, kek_name, dek_expiry_days)
        return transform

    def close(self):
        if self.client is not None:
            self.client.__exit__()

    def _get_cryptor(self, ctx: RuleContext) -> 'Cryptor':
        dek_algorithm = DekAlgorithm.AES256_GCM
        dek_algorithm_str = ctx.get_parameter(ENCRYPT_DEK_ALGORITHM)
        if dek_algorithm_str is not None:
            dek_algorithm = DekAlgorithm[dek_algorithm_str]
        cryptor = Cryptor(dek_algorithm)
        return cryptor

    def _get_kek_name(self, ctx: RuleContext) -> str:
        kek_name = ctx.get_parameter(ENCRYPT_KEK_NAME)
        if kek_name is None:
            raise RuleError("no kek name found")
        if kek_name == "":
            raise RuleError("empty kek name")
        return kek_name

    def _get_dek_expiry_days(self, ctx: RuleContext) -> int:
        dek_expiry_days_str = ctx.get_parameter(ENCRYPT_DEK_EXPIRY_DAYS)
        if dek_expiry_days_str is None:
            return 0
        try:
            dek_expiry_days = int(dek_expiry_days_str)
        except ValueError:
            raise RuleError("invalid expiry days")
        if dek_expiry_days < 0:
            raise RuleError("negative expiry days")
        return dek_expiry_days

    @classmethod
    def register(cls):
        RuleRegistry.register_rule_executor(EncryptionExecutor())

    @classmethod
    def register_with_clock(cls, clock: Clock) -> 'EncryptionExecutor':
        executor = EncryptionExecutor(clock)
        RuleRegistry.register_rule_executor(executor)
        return executor


class Cryptor:
    EMPTY_AAD = b""

    def __init__(self, dek_format: DekAlgorithm):
        self.dek_format = dek_format
        self.is_deterministic = dek_format == DekAlgorithm.AES256_SIV
        self.registry = Registry()

        if dek_format is DekAlgorithm.AES128_GCM:
            self.key_template = aead.aead_key_templates.AES128_GCM_RAW
        elif dek_format is DekAlgorithm.AES256_GCM:
            self.key_template = aead.aead_key_templates.AES256_GCM_RAW
        elif dek_format is DekAlgorithm.AES256_SIV:
            # Construct AES256_SIV_RAW since it doesn't exist in Tink
            key_format = aes_siv_pb2.AesSivKeyFormat(
                # Generate 2 256-bit keys
                key_size=64,
            )
            self.key_template = tink_pb2.KeyTemplate(
                type_url=daead.deterministic_aead_key_templates.AES256_SIV.type_url,
                output_prefix_type=tink_pb2.RAW,
                value=key_format.SerializeToString(),
            )
        else:
            raise RuleError("invalid dek algorithm")

    def generate_key(self) -> bytes:
        key_data = self.registry.new_key_data(self.key_template)
        return key_data.value

    def encrypt(self, dek: bytes, plaintext: bytes, associated_data: bytes) -> bytes:
        key_data = tink_pb2.KeyData(
            type_url=self.key_template.type_url, value=dek, key_material_type=tink_pb2.KeyData.SYMMETRIC
        )
        if self.is_deterministic:
            primitive = self.registry.primitive(key_data, daead.DeterministicAead)
            return primitive.encrypt_deterministically(plaintext, associated_data)
        else:
            primitive = self.registry.primitive(key_data, aead.Aead)
            return primitive.encrypt(plaintext, associated_data)

    def decrypt(self, dek: bytes, ciphertext: bytes, associated_data: bytes) -> bytes:
        key_data = tink_pb2.KeyData(
            type_url=self.key_template.type_url, value=dek, key_material_type=tink_pb2.KeyData.SYMMETRIC
        )
        if self.is_deterministic:
            primitive = self.registry.primitive(key_data, daead.DeterministicAead)
            return primitive.decrypt_deterministically(ciphertext, associated_data)
        else:
            primitive = self.registry.primitive(key_data, aead.Aead)
            return primitive.decrypt(ciphertext, associated_data)


class EncryptionExecutorTransform(object):

    def __init__(self, executor: EncryptionExecutor, cryptor: Cryptor, kek_name: str, dek_expiry_days: int):
        self._executor = executor
        self._cryptor = cryptor
        self._kek_name = kek_name
        self._kek: Optional[Kek] = None
        self._dek_expiry_days = dek_expiry_days

    def _is_dek_rotated(self):
        return self._dek_expiry_days > 0

    def _get_kek(self, ctx: RuleContext) -> Kek:
        if self._kek is None:
            self._kek = self._get_or_create_kek(ctx)
        return self._kek

    def _get_or_create_kek(self, ctx: RuleContext) -> Kek:
        is_read = ctx.rule_mode == RuleMode.READ
        kms_type = ctx.get_parameter(ENCRYPT_KMS_TYPE)
        kms_key_id = ctx.get_parameter(ENCRYPT_KMS_KEY_ID)
        kek_id = KekId(self._kek_name, False)
        kek = self._retrieve_kek_from_registry(kek_id)
        if kek is None:
            if is_read:
                raise RuleError(f"no kek found for {self._kek_name} during consume")
            if not kms_type:
                raise RuleError(f"no kms type found for {self._kek_name} during produce")
            if not kms_key_id:
                raise RuleError(f"no kms key id found for {self._kek_name} during produce")
            kek = self._store_kek_to_registry(kek_id, kms_type, kms_key_id, False)
            if kek is None:
                # handle conflicts (409)
                kek = self._retrieve_kek_from_registry(kek_id)
            if kek is None:
                raise RuleError(f"no kek found for {self._kek_name} during produce")
        if kms_type and kek.kms_type != kms_type:
            raise RuleError(
                f"found {self._kek_name} with kms type {kek.kms_type} " f"which differs from rule kms type {kms_type}"
            )
        if kms_key_id and kek.kms_key_id != kms_key_id:
            raise RuleError(
                f"found {self._kek_name} with kms key id {kek.kms_key_id} "
                f"which differs from rule kms key id {kms_key_id}"
            )
        return kek

    def _retrieve_kek_from_registry(self, kek_id: KekId) -> Optional[Kek]:
        if self._executor.client is None:
            raise RuleError("client not configured")
        try:
            return self._executor.client.get_kek(kek_id.name, kek_id.deleted)
        except Exception as e:
            if isinstance(e, SchemaRegistryError) and e.http_status_code == 404:
                return None
            raise RuleError(f"could not get kek {kek_id.name}") from e

    def _store_kek_to_registry(self, kek_id: KekId, kms_type: str, kms_key_id: str, shared: bool) -> Optional[Kek]:
        if self._executor.client is None:
            raise RuleError("client not configured")
        try:
            return self._executor.client.register_kek(kek_id.name, kms_type, kms_key_id, shared)
        except Exception as e:
            if isinstance(e, SchemaRegistryError) and e.http_status_code == 409:
                return None
            raise RuleError(f"could not register kek {kek_id.name}") from e

    def _get_or_create_dek(self, ctx: RuleContext, version: Optional[int]) -> Dek:
        kek = self._get_kek(ctx)
        is_read = ctx.rule_mode == RuleMode.READ
        if version is None or version == 0:
            version = 1
        # TODO: fallback value for name?
        dek_id = DekId(kek.name, ctx.subject, version, self._cryptor.dek_format, is_read)  # type: ignore[arg-type]
        dek = self._retrieve_dek_from_registry(dek_id)
        is_expired = self._is_expired(ctx, dek)
        primitive = None
        if dek is None or is_expired:
            if is_read:
                raise RuleError(f"no dek found for {dek_id.kek_name} during consume")
            if self._kek is None:
                raise RuleError("no kek found")
            encrypted_dek = None
            if not kek.shared:
                if self._executor.config is None:
                    raise RuleError("config not found in executor")
                primitive = AeadWrapper(self._executor.config, self._kek)
                raw_dek = self._cryptor.generate_key()
                encrypted_dek = primitive.encrypt(raw_dek, self._cryptor.EMPTY_AAD)
            if dek is None or dek.version is None:
                new_version = 1
            else:
                new_version = dek.version + 1 if is_expired else 1
            try:
                dek = self._create_dek(dek_id, new_version, encrypted_dek)
            except RuleError as e:
                if dek is None:
                    raise e
                log.warning(
                    "failed to create dek for %s, subject %s, version %d, using existing dek",
                    kek.name,
                    ctx.subject,
                    new_version,
                )
        key_bytes = dek.get_key_material_bytes()
        if key_bytes is None:
            if primitive is None:
                primitive = AeadWrapper(self._executor.config, self._kek)  # type: ignore[arg-type]
            encrypted_dek = dek.get_encrypted_key_material_bytes()
            raw_dek = primitive.decrypt(encrypted_dek, self._cryptor.EMPTY_AAD)  # type: ignore[arg-type]
            dek.set_key_material(raw_dek)
        return dek

    def _create_dek(self, dek_id: DekId, new_version: Optional[int], encrypted_dek: Optional[bytes]) -> Dek:
        # TODO: fallback value for version?
        new_dek_id = DekId(
            dek_id.kek_name,
            dek_id.subject,
            new_version,  # type: ignore[arg-type]
            dek_id.algorithm,
            dek_id.deleted,
        )
        dek = self._store_dek_to_registry(new_dek_id, encrypted_dek)
        if dek is None:
            # handle conflicts (409)
            dek = self._retrieve_dek_from_registry(dek_id)
        if dek is None:
            raise RuleError(f"no dek found for {dek_id.kek_name} during produce")
        return dek

    def _retrieve_dek_from_registry(self, key: DekId) -> Optional[Dek]:
        try:
            version = key.version
            if not version:
                version = 1
            if self._executor.client is None:
                raise RuleError("client not configured")
            dek = self._executor.client.get_dek(key.kek_name, key.subject, key.algorithm, version, key.deleted)
            return dek if dek and dek.encrypted_key_material else None
        except Exception as e:
            if isinstance(e, SchemaRegistryError) and e.http_status_code == 404:
                return None
            raise RuleError(f"could not get dek for kek {key.kek_name}, subject {key.subject}") from e

    def _store_dek_to_registry(self, key: DekId, encrypted_dek: Optional[bytes]) -> Optional[Dek]:
        try:
            encrypted_dek_str = base64.b64encode(encrypted_dek).decode("utf-8") if encrypted_dek else None
            if self._executor.client is None:
                raise RuleError("client not configured")
            dek = self._executor.client.register_dek(
                key.kek_name, key.subject, encrypted_dek_str, key.algorithm, key.version  # type: ignore[arg-type]
            )
            return dek
        except Exception as e:
            if isinstance(e, SchemaRegistryError) and e.http_status_code == 409:
                return None
            raise RuleError(f"could not register dek for kek {key.kek_name}, subject {key.subject}") from e

    def _is_expired(self, ctx: RuleContext, dek: Optional[Dek]) -> bool:
        now = self._executor.clock.now()
        return (
            ctx.rule_mode != RuleMode.READ
            and self._dek_expiry_days > 0
            and dek is not None
            and (now - (dek.ts or 0)) / MILLIS_IN_DAY > self._dek_expiry_days
        )  # type: ignore[operator]

    def transform(self, ctx: RuleContext, field_type: FieldType, field_value: Any) -> Any:
        if field_value is None:
            return None
        if ctx.rule_mode == RuleMode.WRITE:
            plaintext = self._to_bytes(field_type, field_value)
            if plaintext is None:
                raise RuleError(f"type {field_type} not supported for encryption")
            version = None
            if self._is_dek_rotated():
                version = -1
            dek = self._get_or_create_dek(ctx, version)
            key_material_bytes = dek.get_key_material_bytes()
            if key_material_bytes is None:
                raise RuleError("no key material bytes found for dek")
            ciphertext = self._cryptor.encrypt(key_material_bytes, plaintext, Cryptor.EMPTY_AAD)
            if self._is_dek_rotated():
                if dek.version is None:
                    raise RuleError("no version found for dek")
                ciphertext = self._prefix_version(dek.version, ciphertext)
            if field_type == FieldType.STRING:
                return base64.b64encode(ciphertext).decode("utf-8")
            else:
                return self._to_object(field_type, ciphertext)
        elif ctx.rule_mode == RuleMode.READ:
            ciphertext = None
            if field_type == FieldType.STRING:
                ciphertext = base64.b64decode(field_value)
            else:
                ciphertext = self._to_bytes(field_type, field_value)
            if ciphertext is None:
                return field_value

            version = None
            if self._is_dek_rotated():
                version, ciphertext = self._extract_version(ciphertext)
                if version is None:
                    raise RuleError("no version found in ciphertext")
            dek = self._get_or_create_dek(ctx, version)
            key_material_bytes = dek.get_key_material_bytes()
            if key_material_bytes is None:
                raise RuleError("no key material bytes found for dek")
            plaintext = self._cryptor.decrypt(key_material_bytes, ciphertext, Cryptor.EMPTY_AAD)
            return self._to_object(field_type, plaintext)
        else:
            raise RuleError(f"unsupported rule mode {ctx.rule_mode}")

    def _prefix_version(self, version: int, ciphertext: bytes) -> bytes:
        return bytes([_MAGIC_BYTE_V0]) + version.to_bytes(4, byteorder="big") + ciphertext

    def _extract_version(self, ciphertext: bytes) -> Tuple[Optional[int], bytes]:
        if len(ciphertext) < 5:
            return None, ciphertext
        version = int.from_bytes(ciphertext[1:5], byteorder="big")
        return version, ciphertext[5:]

    def _to_bytes(self, field_type: FieldType, value: Any) -> Optional[bytes]:
        if field_type == FieldType.STRING:
            return value.encode("utf-8")
        elif field_type == FieldType.BYTES:
            if isinstance(value, io.BytesIO):
                return value.read()
            return value
        return None

    def _to_object(self, field_type: FieldType, value: bytes) -> Any:
        if field_type == FieldType.STRING:
            return value.decode("utf-8")
        elif field_type == FieldType.BYTES:
            return value
        return None


class AeadWrapper(aead.Aead):
    def __init__(self, config: dict, kek: Kek):
        self._config = config
        self._kek = kek
        self._kms_key_ids = self._get_kms_key_ids()

    def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
        for index, kms_key_id in enumerate(self._kms_key_ids):
            try:
                if self._kek.kms_type is None:
                    raise RuleError("no kms type found for kek")
                aead = self._get_aead(self._config, self._kek.kms_type, kms_key_id)
                return aead.encrypt(plaintext, associated_data)
            except Exception as e:
                log.warning("failed to encrypt with kek %s and kms key id %s", self._kek.name, kms_key_id)
                if index == len(self._kms_key_ids) - 1:
                    raise RuleError(f"failed to encrypt with all KEKs for {self._kek.name}") from e
        raise RuleError("No KEK found for encryption")

    def decrypt(self, ciphertext: bytes, associated_data: bytes) -> bytes:
        for index, kms_key_id in enumerate(self._kms_key_ids):
            try:
                if self._kek.kms_type is None:
                    raise RuleError("no kms type found for kek")
                aead = self._get_aead(self._config, self._kek.kms_type, kms_key_id)
                return aead.decrypt(ciphertext, associated_data)
            except Exception as e:
                log.warning("failed to decrypt with kek %s and kms key id %s", self._kek.name, kms_key_id)
                if index == len(self._kms_key_ids) - 1:
                    raise RuleError(f"failed to decrypt with all KEKs for {self._kek.name}") from e
        raise RuleError("No KEK found for decryption")

    def _get_kms_key_ids(self) -> List[str]:
        kms_key_ids = [self._kek.kms_key_id]
        alternate_kms_key_ids = None
        if self._kek.kms_props is not None:
            alternate_kms_key_ids = self._kek.kms_props.properties.get(ENCRYPT_ALTERNATE_KMS_KEY_IDS)
        if alternate_kms_key_ids is None:
            alternate_kms_key_ids = self._config.get(ENCRYPT_ALTERNATE_KMS_KEY_IDS)
        if alternate_kms_key_ids is not None:
            # Split the comma-separated list of alternate KMS key IDs and append to kms_key_ids
            kms_key_ids.extend([id.strip() for id in alternate_kms_key_ids.split(',') if id.strip()])
        return kms_key_ids  # type: ignore[return-value]

    def _get_aead(self, config: dict, kms_type: str, kms_key_id: str) -> aead.Aead:
        kek_url = kms_type + "://" + kms_key_id
        kms_client = self._get_kms_client(config, kek_url)
        return kms_client.get_aead(kek_url)

    def _get_kms_client(self, config: dict, kek_url: str) -> KmsClient:
        driver = get_kms_driver(kek_url)
        try:
            client = kms_client_from_uri(kek_url)
        except TinkError:
            client = self._register_kms_client(driver, config, kek_url)
        return client

    def _register_kms_client(self, kms_driver: KmsDriver, config: dict, kek_url: str) -> KmsClient:
        kms_client = kms_driver.new_kms_client(config, kek_url)
        register_kms_client(kms_client)
        return kms_client


class FieldEncryptionExecutor(FieldRuleExecutor):

    def __init__(self, clock: Clock = Clock()):
        self.executor = EncryptionExecutor(clock)

    def configure(self, client_conf: dict, rule_conf: dict):
        self.executor.configure(client_conf, rule_conf)

    def type(self) -> str:
        return "ENCRYPT"

    def new_transform(self, ctx: RuleContext) -> FieldTransform:
        executor_transform = self.executor.new_transform(ctx)
        transform = FieldEncryptionExecutorTransform(executor_transform)
        return transform.transform

    def close(self):
        if self.client is not None:
            self.client.__exit__()

    @classmethod
    def register(cls):
        RuleRegistry.register_rule_executor(FieldEncryptionExecutor())

    @classmethod
    def register_with_clock(cls, clock: Clock) -> 'FieldEncryptionExecutor':
        executor = FieldEncryptionExecutor(clock)
        RuleRegistry.register_rule_executor(executor)
        return executor


class FieldEncryptionExecutorTransform(object):

    def __init__(self, executor_transform: 'EncryptionExecutorTransform'):
        self.executor_transform = executor_transform

    def transform(self, ctx: RuleContext, field_ctx: FieldContext, field_value: Any) -> Any:
        return self.executor_transform.transform(ctx, field_ctx.field_type, field_value)



# Copyright 2023 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

from .._model import TopicPartitionInfo
from .._util import ConversionUtil
from ..cimpl import Uuid
from ._acl import AclOperation


class TopicDescription:
    """
    Represents topic description information for a topic used in describe topic operation.
    Used by :meth:`AdminClient.describe_topics`.

    Parameters
    ----------
    name : str
        The topic name.
    topic_id: Uuid
        The topic id of the topic
    is_internal:
        Whether the topic is internal or not
    partitions : list(TopicPartitionInfo)
        Partition information.
    authorized_operations: list(AclOperation)
        AclOperations allowed for the topic.
    """

    def __init__(
        self,
        name: str,
        topic_id: Uuid,
        is_internal: bool,
        partitions: List[TopicPartitionInfo],
        authorized_operations: Optional[List[Union[str, int, AclOperation]]] = None,
    ) -> None:
        self.name = name
        self.topic_id = topic_id
        self.is_internal = is_internal
        self.partitions = partitions
        self.authorized_operations = None
        if authorized_operations:
            self.authorized_operations = []
            for op in authorized_operations:
                self.authorized_operations.append(ConversionUtil.convert_to_enum(op, AclOperation))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020-2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from typing import Any, Callable, List, Optional, Set, Tuple, Union, cast

from google.protobuf import descriptor_pb2, json_format
from google.protobuf.descriptor import Descriptor, FileDescriptor
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.message import DecodeError, Message
from google.protobuf.message_factory import GetMessageClass

from confluent_kafka.schema_registry import (
    RuleMode,
    Schema,
    SchemaReference,
    dual_schema_id_deserializer,
    prefix_schema_id_serializer,
    reference_subject_name_strategy,
    topic_subject_name_strategy,
)
from confluent_kafka.schema_registry.common.protobuf import (
    PROTOBUF_TYPE,
    _bytes,
    _ContextStringIO,
    _create_index_array,
    _init_pool,
    _is_builtin,
    _schema_to_str,
    _str_to_proto,
    transform,
)
from confluent_kafka.schema_registry.common.schema_registry_client import RulePhase
from confluent_kafka.schema_registry.rule_registry import RuleRegistry
from confluent_kafka.schema_registry.schema_registry_client import SchemaRegistryClient
from confluent_kafka.schema_registry.serde import (
    BaseDeserializer,
    BaseSerializer,
    ParsedSchemaCache,
    SchemaId,
)
from confluent_kafka.serialization import SerializationContext, SerializationError

__all__ = [
    '_resolve_named_schema',
    'ProtobufSerializer',
    'ProtobufDeserializer',
]


def _resolve_named_schema(
    schema: Schema,
    schema_registry_client: SchemaRegistryClient,
    pool: DescriptorPool,
    visited: Optional[Set[str]] = None,
):
    """
    Resolves named schemas referenced by the provided schema recursively.

    :param schema: Schema to resolve named schemas for.
    :param schema_registry_client: SchemaRegistryClient to use for retrieval.
    :param pool: DescriptorPool to add resolved schemas to.
    :return: DescriptorPool
    """
    if visited is None:
        visited = set()
    if schema.references is not None:
        for ref in schema.references:
            if ref.name is None:
                raise ValueError("Name cannot be None")

            if _is_builtin(ref.name) or ref.name in visited:
                continue
            visited.add(ref.name)

            if ref.subject is None or ref.version is None:
                raise ValueError("Subject or version cannot be None")
            referenced_schema = schema_registry_client.get_version(ref.subject, ref.version, True, 'serialized')
            if referenced_schema.schema.schema_str is None:
                raise ValueError("Schema string cannot be None")
            _resolve_named_schema(referenced_schema.schema, schema_registry_client, pool, visited)
            file_descriptor_proto = _str_to_proto(ref.name, referenced_schema.schema.schema_str)
            pool.Add(file_descriptor_proto)


class ProtobufSerializer(BaseSerializer):
    """
    Serializer for Protobuf Message derived classes. Serialization format is Protobuf,
    with Confluent Schema Registry framing.

    Configuration properties:

    +-------------------------------------+----------+------------------------------------------------------+
    | Property Name                       | Type     | Description                                          |
    +=====================================+==========+======================================================+
    |                                     |          | If True, automatically register the configured       |
    | ``auto.register.schemas``           | bool     | schema with Confluent Schema Registry if it has      |
    |                                     |          | not previously been associated with the relevant     |
    |                                     |          | subject (determined via subject.name.strategy).      |
    |                                     |          |                                                      |
    |                                     |          | Defaults to True.                                    |
    |                                     |          |                                                      |
    |                                     |          | Raises SchemaRegistryError if the schema was not     |
    |                                     |          | registered against the subject, or could not be      |
    |                                     |          | successfully registered.                             |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Whether to normalize schemas, which will             |
    | ``normalize.schemas``               | bool     | transform schemas to have a consistent format,       |
    |                                     |          | including ordering properties and references.        |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Whether to use the given schema ID for               |
    | ``use.schema.id``                   | int      | serialization.                                       |
    |                                     |          |                                                      |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Whether to use the latest subject version for        |
    | ``use.latest.version``              | bool     | serialization.                                       |
    |                                     |          |                                                      |
    |                                     |          | WARNING: There is no check that the latest           |
    |                                     |          | schema is backwards compatible with the object       |
    |                                     |          | being serialized.                                    |
    |                                     |          |                                                      |
    |                                     |          | Defaults to False.                                   |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Whether to use the latest subject version with       |
    | ``use.latest.with.metadata``        | dict     | the given metadata.                                  |
    |                                     |          |                                                      |
    |                                     |          | WARNING: There is no check that the latest           |
    |                                     |          | schema is backwards compatible with the object       |
    |                                     |          | being serialized.                                    |
    |                                     |          |                                                      |
    |                                     |          | Defaults to None.                                    |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Whether or not to skip known types when resolving    |
    | ``skip.known.types``                | bool     | schema dependencies.                                 |
    |                                     |          |                                                      |
    |                                     |          | Defaults to True.                                    |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Callable(SerializationContext, str) -> str           |
    |                                     |          |                                                      |
    | ``subject.name.strategy``           | callable | Defines how Schema Registry subject names are        |
    |                                     |          | constructed. Standard naming strategies are          |
    |                                     |          | defined in the confluent_kafka.schema_registry       |
    |                                     |          | namespace.                                           |
    |                                     |          |                                                      |
    |                                     |          | Defaults to topic_subject_name_strategy.             |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Callable(SerializationContext, str) -> str           |
    |                                     |          |                                                      |
    | ``reference.subject.name.strategy`` | callable | Defines how Schema Registry subject names for schema |
    |                                     |          | references are constructed.                          |
    |                                     |          |                                                      |
    |                                     |          | Defaults to reference_subject_name_strategy          |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Callable(bytes, SerializationContext, schema_id)     |
    |                                     |          |   -> bytes                                           |
    |                                     |          |                                                      |
    | ``schema.id.serializer``            | callable | Defines how the schema id/guid is serialized.        |
    |                                     |          | Defaults to prefix_schema_id_serializer.             |
    +-------------------------------------+----------+------------------------------------------------------+

    Schemas are registered against subject names in Confluent Schema Registry that
    define a scope in which the schemas can be evolved. By default, the subject name
    is formed by concatenating the topic name with the message field (key or value)
    separated by a hyphen.

    i.e. {topic name}-{message field}

    Alternative naming strategies may be configured with the property
    ``subject.name.strategy``.

    Supported subject name strategies

    +--------------------------------------+------------------------------+
    | Subject Name Strategy                | Output Format                |
    +======================================+==============================+
    | topic_subject_name_strategy(default) | {topic name}-{message field} |
    +--------------------------------------+------------------------------+
    | topic_record_subject_name_strategy   | {topic name}-{record name}   |
    +--------------------------------------+------------------------------+
    | record_subject_name_strategy         | {record name}                |
    +--------------------------------------+------------------------------+

    See `Subject name strategy <https://docs.confluent.io/current/schema-registry/serializer-formatter.html#subject-name-strategy>`_ for additional details.

    Args:
        msg_type (Message): Protobuf Message type.

        schema_registry_client (SchemaRegistryClient): Schema Registry
            client instance.

        conf (dict): ProtobufSerializer configuration.

    See Also:
        `Protobuf API reference <https://googleapis.dev/python/protobuf/latest/google/protobuf.html>`_
    """  # noqa: E501

    __slots__ = [
        '_skip_known_types',
        '_known_subjects',
        '_msg_class',
        '_index_array',
        '_schema',
        '_schema_id',
        '_ref_reference_subject_func',
        '_use_deprecated_format',
        '_parsed_schemas',
    ]

    _default_conf = {
        'auto.register.schemas': True,
        'normalize.schemas': False,
        'use.schema.id': None,
        'use.latest.version': False,
        'use.latest.with.metadata': None,
        'skip.known.types': True,
        'subject.name.strategy': topic_subject_name_strategy,
        'reference.subject.name.strategy': reference_subject_name_strategy,
        'schema.id.serializer': prefix_schema_id_serializer,
        'use.deprecated.format': False,
    }

    def __init_impl(
        self,
        msg_type: Message,
        schema_registry_client: SchemaRegistryClient,
        conf: Optional[dict] = None,
        rule_conf: Optional[dict] = None,
        rule_registry: Optional[RuleRegistry] = None,
    ):
        super().__init__()

        conf_copy = self._default_conf.copy()
        if conf is not None:
            conf_copy.update(conf)

        self._auto_register = cast(bool, conf_copy.pop('auto.register.schemas'))
        if not isinstance(self._auto_register, bool):
            raise ValueError("auto.register.schemas must be a boolean value")

        self._normalize_schemas = cast(bool, conf_copy.pop('normalize.schemas'))
        if not isinstance(self._normalize_schemas, bool):
            raise ValueError("normalize.schemas must be a boolean value")

        self._use_schema_id = cast(Optional[int], conf_copy.pop('use.schema.id'))
        if self._use_schema_id is not None and not isinstance(self._use_schema_id, int):
            raise ValueError("use.schema.id must be an int value")

        self._use_latest_version = cast(bool, conf_copy.pop('use.latest.version'))
        if not isinstance(self._use_latest_version, bool):
            raise ValueError("use.latest.version must be a boolean value")
        if self._use_latest_version and self._auto_register:
            raise ValueError("cannot enable both use.latest.version and auto.register.schemas")

        self._use_latest_with_metadata = cast(Optional[dict], conf_copy.pop('use.latest.with.metadata'))
        if self._use_latest_with_metadata is not None and not isinstance(self._use_latest_with_metadata, dict):
            raise ValueError("use.latest.with.metadata must be a dict value")

        self._skip_known_types = cast(bool, conf_copy.pop('skip.known.types'))
        if not isinstance(self._skip_known_types, bool):
            raise ValueError("skip.known.types must be a boolean value")

        self._use_deprecated_format = cast(bool, conf_copy.pop('use.deprecated.format'))
        if not isinstance(self._use_deprecated_format, bool):
            raise ValueError("use.deprecated.format must be a boolean value")
        if self._use_deprecated_format:
            raise ValueError("use.deprecated.format is no longer supported")

        self._subject_name_func = cast(
            Callable[[Optional[SerializationContext], Optional[str]], Optional[str]],
            conf_copy.pop('subject.name.strategy'),
        )
        if not callable(self._subject_name_func):
            raise ValueError("subject.name.strategy must be callable")

        self._ref_reference_subject_func = cast(
            Callable[[Optional[SerializationContext], Any], Optional[str]],
            conf_copy.pop('reference.subject.name.strategy'),
        )
        if not callable(self._ref_reference_subject_func):
            raise ValueError("subject.name.strategy must be callable")

        self._schema_id_serializer = cast(
            Callable[[bytes, Optional[SerializationContext], Any], bytes], conf_copy.pop('schema.id.serializer')
        )
        if not callable(self._schema_id_serializer):
            raise ValueError("schema.id.serializer must be callable")

        if len(conf_copy) > 0:
            raise ValueError("Unrecognized properties: {}".format(", ".join(conf_copy.keys())))

        self._registry = schema_registry_client
        self._rule_registry = rule_registry if rule_registry else RuleRegistry.get_global_instance()
        self._schema_id: Optional[SchemaId] = None
        self._known_subjects: set[str] = set()
        self._msg_class = msg_type
        self._parsed_schemas = ParsedSchemaCache()

        descriptor = msg_type.DESCRIPTOR
        self._index_array = _create_index_array(descriptor)
        self._schema = Schema(_schema_to_str(descriptor.file), schema_type='PROTOBUF')

        for rule in self._rule_registry.get_executors():
            rule.configure(self._registry.config() if self._registry else {}, rule_conf if rule_conf else {})

    __init__ = __init_impl

    @staticmethod
    def _write_varint(buf: io.BytesIO, val: int, zigzag: bool = True):
        """
        Writes val to buf, either using zigzag or uvarint encoding.

        Args:
            buf (BytesIO): buffer to write to.
            val (int): integer to be encoded.
            zigzag (bool): whether to encode in zigzag or uvarint encoding
        """

        if zigzag:
            val = (val << 1) ^ (val >> 63)

        while (val & ~0x7F) != 0:
            buf.write(_bytes((val & 0x7F) | 0x80))
            val >>= 7
        buf.write(_bytes(val))

    @staticmethod
    def _encode_varints(buf: io.BytesIO, ints: List[int], zigzag: bool = True):
        """
        Encodes each int as a uvarint onto buf

        Args:
            buf (BytesIO): buffer to write to.
            ints ([int]): ints to be encoded.
            zigzag (bool): whether to encode in zigzag or uvarint encoding
        """

        assert len(ints) > 0
        # The root element at the 0 position does not need a length prefix.
        if ints == [0]:
            buf.write(_bytes(0x00))
            return

        ProtobufSerializer._write_varint(buf, len(ints), zigzag=zigzag)

        for value in ints:
            ProtobufSerializer._write_varint(buf, value, zigzag=zigzag)

    def _resolve_dependencies(self, ctx: SerializationContext, file_desc: FileDescriptor) -> List[SchemaReference]:
        """
        Resolves and optionally registers schema references recursively.

        Args:
            ctx (SerializationContext): Serialization context.

            file_desc (FileDescriptor): file descriptor to traverse.
        """

        schema_refs = []
        for dep in file_desc.dependencies:
            if self._skip_known_types and _is_builtin(dep.name):
                continue
            dep_refs = self._resolve_dependencies(ctx, dep)
            subject = self._ref_reference_subject_func(ctx, dep)
            schema = Schema(_schema_to_str(dep), references=dep_refs, schema_type='PROTOBUF')
            if self._auto_register:
                self._registry.register_schema(subject, schema)

            reference = self._registry.lookup_schema(subject, schema)
            # schema_refs are per file descriptor
            schema_refs.append(SchemaReference(dep.name, subject, reference.version))
        return schema_refs

    def __call__(  # type: ignore[override]
        self, message: Message, ctx: Optional[SerializationContext] = None
    ) -> Optional[bytes]:
        return self.__serialize(message, ctx)

    def __serialize(self, message: Message, ctx: Optional[SerializationContext] = None) -> Optional[bytes]:
        """
        Serializes an instance of a class derived from Protobuf Message, and prepends
        it with Confluent Schema Registry framing.

        Args:
            message (Message): An instance of a class derived from Protobuf Message.

            ctx (SerializationContext): Metadata relevant to the serialization.
                operation.

        Raises:
            SerializerError if any error occurs during serialization.

        Returns:
            None if messages is None, else a byte array containing the Protobuf
            serialized message with Confluent Schema Registry framing.
        """

        if message is None:
            return None

        if not isinstance(message, self._msg_class):
            raise ValueError("message must be of type {} not {}".format(self._msg_class, type(message)))

        subject = self._subject_name_func(ctx, message.DESCRIPTOR.full_name) if ctx else None
        latest_schema = None
        if subject is not None:
            latest_schema = self._get_reader_schema(subject, fmt='serialized')

        if latest_schema is not None:
            self._schema_id = SchemaId(PROTOBUF_TYPE, latest_schema.schema_id, latest_schema.guid, self._index_array)

        elif subject is not None and subject not in self._known_subjects and ctx is not None:
            references = self._resolve_dependencies(ctx, message.DESCRIPTOR.file)
            self._schema = Schema(self._schema.schema_str, self._schema.schema_type, references)

            if self._auto_register:
                registered_schema = self._registry.register_schema_full_response(
                    subject, self._schema, normalize_schemas=self._normalize_schemas
                )
                self._schema_id = SchemaId(
                    PROTOBUF_TYPE, registered_schema.schema_id, registered_schema.guid, self._index_array
                )
            else:
                registered_schema = self._registry.lookup_schema(
                    subject, self._schema, normalize_schemas=self._normalize_schemas
                )
                self._schema_id = SchemaId(
                    PROTOBUF_TYPE, registered_schema.schema_id, registered_schema.guid, self._index_array
                )

            self._known_subjects.add(subject)

        if latest_schema is not None:
            fd_proto, pool = self._get_parsed_schema(latest_schema.schema)
            fd = pool.FindFileByName(fd_proto.name)
            desc = fd.message_types_by_name[message.DESCRIPTOR.name]

            def field_transformer(rule_ctx, field_transform, msg):
                return transform(rule_ctx, desc, msg, field_transform)  # noqa: E731

            if ctx is not None and subject is not None:
                message = self._execute_rules(
                    ctx, subject, RuleMode.WRITE, None, latest_schema.schema, message, None, field_transformer
                )

        with _ContextStringIO() as fo:
            fo.write(message.SerializeToString())
            if self._schema_id is not None:
                self._schema_id.message_indexes = self._index_array
            buffer = fo.getvalue()

            if latest_schema is not None and ctx is not None and subject is not None:
                buffer = self._execute_rules_with_phase(
                    ctx, subject, RulePhase.ENCODING, RuleMode.WRITE, None, latest_schema.schema, buffer, None, None
                )

            return self._schema_id_serializer(buffer, ctx, self._schema_id)

    def _get_parsed_schema(self, schema: Schema) -> Tuple[descriptor_pb2.FileDescriptorProto, DescriptorPool]:
        result = self._parsed_schemas.get_parsed_schema(schema)
        if result is not None:
            return result

        pool = DescriptorPool()
        _init_pool(pool)
        _resolve_named_schema(schema, self._registry, pool)
        if schema.schema_str is None:
            raise ValueError("Schema string cannot be None")
        fd_proto = _str_to_proto("default", schema.schema_str)
        pool.Add(fd_proto)
        self._parsed_schemas.set(schema, (fd_proto, pool))
        return fd_proto, pool


class ProtobufDeserializer(BaseDeserializer):
    """
    Deserializer for Protobuf serialized data with Confluent Schema Registry framing.

    Args:
        message_type (Message derived type): Protobuf Message type.
        conf (dict): Configuration dictionary.

    ProtobufDeserializer configuration properties:

    +-------------------------------------+----------+------------------------------------------------------+
    | Property Name                       | Type     | Description                                          |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Whether to use the latest subject version for        |
    | ``use.latest.version``              | bool     | deserialization.                                     |
    |                                     |          |                                                      |
    |                                     |          | Defaults to False.                                   |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Whether to use the latest subject version with       |
    | ``use.latest.with.metadata``        | dict     | the given metadata.                                  |
    |                                     |          |                                                      |
    |                                     |          | Defaults to None.                                    |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Callable(SerializationContext, str) -> str           |
    |                                     |          |                                                      |
    | ``subject.name.strategy``           | callable | Defines how Schema Registry subject names are        |
    |                                     |          | constructed. Standard naming strategies     are      |
    |                                     |          | defined in the confluent_kafka.    schema_registry   |
    |                                     |          | namespace    .                                       |
    |                                     |          |                                                      |
    |                                     |          | Defaults to topic_subject_name_strategy.             |
    +-------------------------------------+----------+------------------------------------------------------+
    |                                     |          | Callable(bytes, SerializationContext, schema_id)     |
    |                                     |          |   -> io.BytesIO                                      |
    |                                     |          |                                                      |
    | ``schema.id.deserializer``          | callable | Defines how the schema id/guid is deserialized.      |
    |                                     |          | Defaults to dual_schema_id_deserializer.             |
    +-------------------------------------+----------+------------------------------------------------------+


    See Also:
    `Protobuf API reference <https://googleapis.dev/python/protobuf/latest/google/protobuf.html>`_
    """

    __slots__ = ['_msg_class', '_use_deprecated_format', '_parsed_schemas']

    _default_conf = {
        'use.latest.version': False,
        'use.latest.with.metadata': None,
        'subject.name.strategy': topic_subject_name_strategy,
        'schema.id.deserializer': dual_schema_id_deserializer,
        'use.deprecated.format': False,
    }

    def __init_impl(
        self,
        message_type: Message,
        conf: Optional[dict] = None,
        schema_registry_client: Optional[SchemaRegistryClient] = None,
        rule_conf: Optional[dict] = None,
        rule_registry: Optional[RuleRegistry] = None,
    ):
        super().__init__()

        self._registry = schema_registry_client
        self._rule_registry = rule_registry if rule_registry else RuleRegistry.get_global_instance()
        self._parsed_schemas = ParsedSchemaCache()
        self._use_schema_id = None

        conf_copy = self._default_conf.copy()
        if conf is not None:
            conf_copy.update(conf)

        self._use_latest_version = cast(bool, conf_copy.pop('use.latest.version'))
        if not isinstance(self._use_latest_version, bool):
            raise ValueError("use.latest.version must be a boolean value")

        self._use_latest_with_metadata = cast(Optional[dict], conf_copy.pop('use.latest.with.metadata'))
        if self._use_latest_with_metadata is not None and not isinstance(self._use_latest_with_metadata, dict):
            raise ValueError("use.latest.with.metadata must be a dict value")

        self._subject_name_func = cast(
            Callable[[Optional[SerializationContext], Optional[str]], Optional[str]],
            conf_copy.pop('subject.name.strategy'),
        )
        if not callable(self._subject_name_func):
            raise ValueError("subject.name.strategy must be callable")

        self._schema_id_deserializer = cast(
            Callable[[bytes, Optional[SerializationContext], Any], io.BytesIO], conf_copy.pop('schema.id.deserializer')
        )
        if not callable(self._schema_id_deserializer):
            raise ValueError("schema.id.deserializer must be callable")

        self._use_deprecated_format = cast(bool, conf_copy.pop('use.deprecated.format'))
        if not isinstance(self._use_deprecated_format, bool):
            raise ValueError("use.deprecated.format must be a boolean value")
        if self._use_deprecated_format:
            raise ValueError("use.deprecated.format is no longer supported")

        descriptor = message_type.DESCRIPTOR
        self._msg_class = GetMessageClass(descriptor)

        for rule in self._rule_registry.get_executors():
            rule.configure(self._registry.config() if self._registry else {}, rule_conf if rule_conf else {})

    __init__ = __init_impl

    def __call__(self, data: Optional[bytes], ctx: Optional[SerializationContext] = None) -> Optional[bytes]:
        return self.__deserialize(data, ctx)

    def __deserialize(self, data: Optional[bytes], ctx: Optional[SerializationContext] = None) -> Optional[bytes]:
        """
        Deserialize a serialized protobuf message with Confluent Schema Registry
        framing.

        Args:
            data (bytes): Serialized protobuf message with Confluent Schema
                           Registry framing.

            ctx (SerializationContext): Metadata relevant to the serialization
                operation.

        Returns:
            Message: Protobuf Message instance.

        Raises:
            SerializerError: If there was an error reading the Confluent framing
                data, or parsing the protobuf serialized message.
        """

        if data is None:
            return None

        subject = self._subject_name_func(ctx, None)
        latest_schema = None
        if subject is not None and self._registry is not None:
            latest_schema = self._get_reader_schema(subject, fmt='serialized')

        schema_id = SchemaId(PROTOBUF_TYPE)
        payload = self._schema_id_deserializer(data, ctx, schema_id)
        msg_index = schema_id.message_indexes

        if self._registry is not None:
            writer_schema_raw = self._get_writer_schema(schema_id, subject, fmt='serialized')
            fd_proto, pool = self._get_parsed_schema(writer_schema_raw)
            writer_schema = pool.FindFileByName(fd_proto.name)
            writer_desc = self._get_message_desc(pool, writer_schema, msg_index if msg_index is not None else [])
            if subject is None:
                subject = self._subject_name_func(ctx, writer_desc.full_name)
                if subject is not None:
                    latest_schema = self._get_reader_schema(subject, fmt='serialized')
        else:
            writer_schema_raw = None
            writer_schema = None

        if ctx is not None and subject is not None:
            payload = self._execute_rules_with_phase(
                ctx, subject, RulePhase.ENCODING, RuleMode.READ, None, writer_schema_raw, payload, None, None
            )
        if isinstance(payload, bytes):
            payload = io.BytesIO(payload)

        reader_schema_raw: Optional[Schema] = None
        if latest_schema is not None and subject is not None and writer_schema_raw is not None:
            migrations = self._get_migrations(subject, writer_schema_raw, latest_schema, None)
            reader_schema_raw = latest_schema.schema
            fd_proto, pool = self._get_parsed_schema(latest_schema.schema)
            reader_schema = pool.FindFileByName(fd_proto.name)
        else:
            migrations = None
            reader_schema_raw = writer_schema_raw
            reader_schema = writer_schema

        if reader_schema is not None:
            # Initialize reader desc to first message in file
            reader_desc = self._get_message_desc(pool, reader_schema, [0])
            # Attempt to find a reader desc with the same name as the writer
            reader_desc = reader_schema.message_types_by_name.get(writer_desc.name, reader_desc)

        if migrations and ctx is not None and subject is not None:
            msg = GetMessageClass(writer_desc)()
            try:
                msg.ParseFromString(payload.read())
            except DecodeError as e:
                raise SerializationError(str(e))

            obj_dict = json_format.MessageToDict(msg, True)
            obj_dict = self._execute_migrations(ctx, subject, migrations, obj_dict)
            msg = GetMessageClass(reader_desc)()
            msg = json_format.ParseDict(obj_dict, msg)
        else:
            # Protobuf Messages are self-describing; no need to query schema
            msg = self._msg_class()
            try:
                msg.ParseFromString(payload.read())
            except DecodeError as e:
                raise SerializationError(str(e))

        def field_transformer(rule_ctx, field_transform, message):
            return transform(rule_ctx, reader_desc, message, field_transform)  # noqa: E731

        if ctx is not None and subject is not None:
            msg = self._execute_rules(
                ctx, subject, RuleMode.READ, None, reader_schema_raw, msg, None, field_transformer
            )
        return msg

    def _get_parsed_schema(self, schema: Schema) -> Tuple[descriptor_pb2.FileDescriptorProto, DescriptorPool]:
        result = self._parsed_schemas.get_parsed_schema(schema)
        if result is not None:
            return result

        pool = DescriptorPool()
        _init_pool(pool)
        _resolve_named_schema(schema, self._registry, pool)
        if schema.schema_str is None:
            raise ValueError("Schema string cannot be None")
        fd_proto = _str_to_proto("default", schema.schema_str)
        pool.Add(fd_proto)
        self._parsed_schemas.set(schema, (fd_proto, pool))
        return fd_proto, pool

    def _get_message_desc(self, pool: DescriptorPool, fd: FileDescriptor, msg_index: List[int]) -> Descriptor:
        file_desc_proto = descriptor_pb2.FileDescriptorProto()
        fd.CopyToProto(file_desc_proto)
        full_name, desc_proto = self._get_message_desc_proto("", file_desc_proto, msg_index)
        package = file_desc_proto.package
        qualified_name = package + "." + full_name if package else full_name
        return pool.FindMessageTypeByName(qualified_name)

    def _get_message_desc_proto(
        self,
        path: str,
        desc: Union[descriptor_pb2.FileDescriptorProto, descriptor_pb2.DescriptorProto],
        msg_index: List[int],
    ) -> Tuple[str, descriptor_pb2.DescriptorProto]:
        index = msg_index[0]
        if isinstance(desc, descriptor_pb2.FileDescriptorProto):
            msg = desc.message_type[index]
            path = path + "." + msg.name if path else msg.name
            if len(msg_index) == 1:
                return path, msg
            return self._get_message_desc_proto(path, msg, msg_index[1:])
        else:
            msg = desc.nested_type[index]
            path = path + "." + msg.name if path else msg.name
            if len(msg_index) == 1:
                return path, msg
            return self._get_message_desc_proto(path, msg, msg_index[1:])



# Copyright 2025 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures
import logging
from typing import Any, Callable, Dict, Optional

import confluent_kafka

from .. import _common as _common
from ._buffer_timeout_manager import BufferTimeoutManager
from ._kafka_batch_executor import ProducerBatchExecutor
from ._producer_batch_processor import ProducerBatchManager

logger = logging.getLogger(__name__)


class AIOProducer:

    # ========================================================================
    # INITIALIZATION AND LIFECYCLE MANAGEMENT
    # ========================================================================

    def __init__(
        self,
        producer_conf: Dict[str, Any],
        max_workers: int = 4,
        executor: Optional[concurrent.futures.Executor] = None,
        batch_size: int = 1000,
        buffer_timeout: float = 1.0,
    ) -> None:
        if executor is not None:
            self.executor = executor
        else:
            self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        # Store the event loop for async operations
        self._loop = asyncio.get_running_loop()

        wrap_common_callbacks = _common.wrap_common_callbacks
        wrap_common_callbacks(self._loop, producer_conf)

        self._producer: confluent_kafka.Producer = confluent_kafka.Producer(producer_conf)

        # Batching configuration
        self._batch_size: int = batch_size

        # Producer state management
        self._is_closed: bool = False  # Track if producer is closed

        # Initialize Kafka batch executor for handling Kafka operations
        self._kafka_executor = ProducerBatchExecutor(self._producer, self.executor)

        # Initialize batch processor for message batching and processing
        self._batch_processor = ProducerBatchManager(self._kafka_executor)

        # Initialize buffer timeout manager for timeout handling
        self._buffer_timeout_manager = BufferTimeoutManager(self._batch_processor, self._kafka_executor, buffer_timeout)
        if buffer_timeout > 0:
            self._buffer_timeout_manager.start_timeout_monitoring()

    async def close(self) -> None:
        """Close the producer and cleanup resources

        This method performs a graceful shutdown sequence to ensure all resources
        are properly cleaned up and no messages are lost:

        1. **Signal Shutdown**: Sets the closed flag to signal the timeout task to stop
        2. **Cancel Timeout Task**: Immediately cancels the buffer timeout monitoring task
        3. **Flush All Messages**: Flushes any buffered messages and waits for delivery confirmation
        4. **Shutdown ThreadPool**: Waits for all pending ThreadPool operations to complete
        5. **Cleanup**: Ensures the underlying librdkafka producer is properly closed. The shutdown
            is designed to be safe and non-blocking for the asyncio event loop
            while ensuring all pending operations complete before the producer is closed.

        Raises:
            Exception: May raise exceptions from buffer flushing, but these are logged
                      and don't prevent the cleanup process from completing.
        """
        # Set closed flag to signal timeout task to stop
        self._is_closed = True

        # Stop the buffer timeout monitoring task
        self._buffer_timeout_manager.stop_timeout_monitoring()

        # Flush any remaining messages
        try:
            await self.flush()
        except Exception:
            logger.error("Error flushing messages during close", exc_info=True)
            raise

        # Shutdown the ThreadPool executor and wait for any remaining tasks to complete
        # This ensures that all pending poll(), flush(), and other blocking operations
        # finish before the producer is considered fully closed
        if hasattr(self, 'executor'):
            # executor.shutdown(wait=True) is a blocking call that:
            # - Prevents new tasks from being submitted to the ThreadPool
            # - Waits for all currently executing and queued tasks to complete
            # - Returns only when all worker threads have finished
            #
            # We run this in a separate thread (using None as executor) to avoid
            # blocking the asyncio event loop during the potentially long shutdown wait
            await asyncio.get_running_loop().run_in_executor(None, self.executor.shutdown, True)

    def __del__(self) -> None:
        """Cleanup method called during garbage collection

        This ensures that the timeout task is properly cancelled even if
        close() wasn't explicitly called.
        """
        if hasattr(self, '_is_closed'):
            self._is_closed = True
        if hasattr(self, '_buffer_timeout_manager'):
            self._buffer_timeout_manager.stop_timeout_monitoring()

    def __len__(self) -> int:
        """Return the total number of pending messages.

        This includes:
        - Messages in librdkafka's output queue (waiting to be delivered to broker)
        - Messages in the async batch buffer (waiting to be sent to librdkafka)

        Returns:
            int: Total number of pending messages across both queues
        """
        if self._is_closed:
            return 0

        # Count messages in librdkafka queue
        librdkafka_count = len(self._producer)

        # Count messages in async batch buffer
        buffer_count = self._batch_processor.get_buffer_size()

        return librdkafka_count + buffer_count

    # ========================================================================
    # CORE PRODUCER OPERATIONS - Main public API
    # ========================================================================

    async def poll(self, timeout: float = 0, *args: Any, **kwargs: Any) -> int:
        """Processes delivery callbacks from librdkafka - blocking depends on timeout

        This method triggers any pending delivery reports that have been
        queued by librdkafka when messages are delivered or fail to deliver.

        Args:
            timeout: Timeout in seconds for waiting for callbacks:
                    - 0 = non-blocking, return after processing available callbacks
                    - >0 = block up to timeout seconds waiting for new callbacks
                    - -1 = block indefinitely until callbacks are available

        Returns:
            Number of callbacks processed during this call
        """
        return await self._call(self._producer.poll, timeout, *args, **kwargs)

    async def produce(
        self, topic: str, value: Optional[Any] = None, key: Optional[Any] = None, *args: Any, **kwargs: Any
    ) -> asyncio.Future[Any]:
        """Batched produce: Accumulates messages in buffer and flushes when threshold reached

        Args:
            topic: Kafka topic name (required)
            value: Message payload (optional)
            key: Message key (optional)
            *args, **kwargs: Additional parameters like partition, timestamp, headers

        Returns:
            asyncio.Future: Future that resolves to the delivered message or raises exception on failure
        """
        result = asyncio.get_running_loop().create_future()

        msg_data = {'topic': topic, 'value': value, 'key': key}

        # Add optional parameters to message data
        if 'partition' in kwargs:
            msg_data['partition'] = kwargs['partition']
        if 'timestamp' in kwargs:
            msg_data['timestamp'] = kwargs['timestamp']
        if 'headers' in kwargs:
            # Headers are not supported in batch mode due to librdkafka API limitations.
            # Use individual synchronous produce() calls if headers are required.
            raise NotImplementedError(
                "Headers are not supported in AIOProducer batch mode. "
                "Use the synchronous Producer.produce() method if headers are required."
            )

        self._batch_processor.add_message(msg_data, result)

        self._buffer_timeout_manager.mark_activity()

        # Check if we should flush the buffer
        if self._batch_processor.get_buffer_size() >= self._batch_size:
            await self._flush_buffer()

        return result

    async def flush(self, *args: Any, **kwargs: Any) -> Any:
        """Waits until all messages are delivered or timeout

        This method performs a complete flush:
        1. Flushes any buffered messages from local buffer to librdkafka
        2. Waits for librdkafka to deliver/acknowledge all messages
        """
        # First, flush any remaining messages in the buffer for all topics
        if not self._batch_processor.is_buffer_empty():
            await self._flush_buffer()
            # Update buffer activity since we just flushed
            self._buffer_timeout_manager.mark_activity()

        # Then flush the underlying producer and wait for delivery confirmation
        return await self._call(self._producer.flush, *args, **kwargs)

    async def purge(self, *args: Any, **kwargs: Any) -> Any:
        """Purges messages from internal queues - may block during cleanup"""
        # Cancel all pending futures
        self._batch_processor.cancel_pending_futures()

        # Clear local message buffer and futures
        self._batch_processor.clear_buffer()

        # Update buffer activity since we cleared the buffer
        self._buffer_timeout_manager.mark_activity()

        return await self._call(self._producer.purge, *args, **kwargs)

    async def list_topics(self, *args: Any, **kwargs: Any) -> Any:
        return await self._call(self._producer.list_topics, *args, **kwargs)

    # ========================================================================
    # TRANSACTION OPERATIONS - Kafka transaction support
    # ========================================================================

    async def init_transactions(self, *args: Any, **kwargs: Any) -> Any:
        """Network call to initialize transactions"""
        return await self._call(self._producer.init_transactions, *args, **kwargs)

    async def begin_transaction(self, *args: Any, **kwargs: Any) -> Any:
        """Network call to begin transaction"""

        # Flush messages to set a clean state before entering a transaction
        await self.flush()

        return await self._call(self._producer.begin_transaction, *args, **kwargs)

    async def send_offsets_to_transaction(self, *args: Any, **kwargs: Any) -> Any:
        """Network call to send offsets to transaction"""
        return await self._call(self._producer.send_offsets_to_transaction, *args, **kwargs)

    async def commit_transaction(self, *args: Any, **kwargs: Any) -> Any:
        """Commit transaction after flushing all buffered messages"""

        # Flush to ensure messages in the local batch_processor buffer are
        # delivered to librdkafka
        await self.flush()

        # Then commit transaction
        return await self._call(self._producer.commit_transaction, *args, **kwargs)

    async def abort_transaction(self, *args: Any, **kwargs: Any) -> Any:
        """Network call to abort transaction

        Messages produced before the call (i.e. inside the transaction boundary) will be aborted.
        Messages that are still in flight may be failed by librdkafka as they are considered
        outside the transaction boundary.
        Refer to librdkafka documentation section "Transactional producer API"
        for more details:
        https://github.com/confluentinc/librdkafka/blob/master/INTRODUCTION.md#transactional-producer
        """

        # Flush to ensure messages in the local batch_processor buffer are
        # delivered to librdkafka
        await self.flush()

        return await self._call(self._producer.abort_transaction, *args, **kwargs)

    # ========================================================================
    # AUTHENTICATION AND SECURITY
    # ========================================================================

    async def set_sasl_credentials(self, *args: Any, **kwargs: Any) -> Any:
        """Authentication operation that may involve network calls"""
        return await self._call(self._producer.set_sasl_credentials, *args, **kwargs)

    # ========================================================================
    # BATCH PROCESSING OPERATIONS - Delegated to BatchProcessor
    # ========================================================================

    async def _flush_buffer(self, target_topic: Optional[str] = None) -> None:
        """Flush the current message buffer using clean batch processing flow

        This method demonstrates the new architecture where AIOProducer simply
        orchestrates the workflow between components:
        1. BatchProcessor creates immutable MessageBatch objects
        2. ProducerBatchExecutor executes each batch
        3. BufferTimeoutManager handles activity tracking
        """
        await self._batch_processor.flush_buffer(target_topic)

    # ========================================================================
    # UTILITY METHODS - Helper functions and internal utilities
    # ========================================================================

    async def _call(self, blocking_task: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        """Helper method for blocking operations that need ThreadPool execution"""
        return await _common.async_call(self.executor, blocking_task, *args, **kwargs)



#!/usr/bin/env python
#
# Copyright 2016-2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Avro schema registry module: Deals with encoding and decoding of messages with avro schemas
"""

import warnings

from confluent_kafka import Consumer, Producer
from confluent_kafka.avro.cached_schema_registry_client import CachedSchemaRegistryClient
from confluent_kafka.avro.error import ClientError
from confluent_kafka.avro.load import load, loads  # noqa
from confluent_kafka.avro.serializer import SerializerError  # noqa
from confluent_kafka.avro.serializer import KeySerializerError, ValueSerializerError
from confluent_kafka.avro.serializer.message_serializer import MessageSerializer


class AvroProducer(Producer):
    """
    .. deprecated:: 2.0.2

    This class will be removed in a future version of the library.

    Kafka Producer client which does avro schema encoding to messages.
    Handles schema registration, Message serialization.

    Constructor arguments:

    :param dict config: Config parameters containing url for schema registry (``schema.registry.url``)
                        and the standard Kafka client configuration (``bootstrap.servers`` et.al).
    :param str default_key_schema: Optional default avro schema for key
    :param str default_value_schema: Optional default avro schema for value
    """

    def __init__(self, config, default_key_schema=None, default_value_schema=None, schema_registry=None, **kwargs):
        warnings.warn(
            "AvroProducer has been deprecated. Use AvroSerializer instead.", category=DeprecationWarning, stacklevel=2
        )

        sr_conf = {
            key.replace("schema.registry.", ""): value
            for key, value in config.items()
            if key.startswith("schema.registry")
        }

        if sr_conf.get("basic.auth.credentials.source") == 'SASL_INHERIT':
            # Fallback to plural 'mechanisms' for backward compatibility
            sr_conf['sasl.mechanism'] = config.get('sasl.mechanism', config.get('sasl.mechanisms', ''))
            sr_conf['sasl.username'] = config.get('sasl.username', '')
            sr_conf['sasl.password'] = config.get('sasl.password', '')
            sr_conf['auto.register.schemas'] = config.get('auto.register.schemas', True)

        ap_conf = {key: value for key, value in config.items() if not key.startswith("schema.registry")}

        if schema_registry is None:
            schema_registry = CachedSchemaRegistryClient(sr_conf)
        elif sr_conf.get("url", None) is not None:
            raise ValueError("Cannot pass schema_registry along with schema.registry.url config")

        super(AvroProducer, self).__init__(ap_conf, **kwargs)
        self._serializer = MessageSerializer(schema_registry)
        self._key_schema = default_key_schema
        self._value_schema = default_value_schema

    def produce(self, **kwargs):
        """
        Asynchronously sends message to Kafka by encoding with specified or default avro schema.

        :param str topic: topic name
        :param object value: An object to serialize
        :param str value_schema: Avro schema for value
        :param object key: An object to serialize
        :param str key_schema: Avro schema for key

        Plus any other parameters accepted by confluent_kafka.Producer.produce

        :raises SerializerError: On serialization failure
        :raises BufferError: If producer queue is full.
        :raises KafkaException: For other produce failures.
        """
        # get schemas from  kwargs if defined
        key_schema = kwargs.pop('key_schema', self._key_schema)
        value_schema = kwargs.pop('value_schema', self._value_schema)
        topic = kwargs.pop('topic', None)
        if not topic:
            raise ClientError("Topic name not specified.")
        value = kwargs.pop('value', None)
        key = kwargs.pop('key', None)

        if value is not None:
            if value_schema:
                value = self._serializer.encode_record_with_schema(topic, value_schema, value)
            else:
                raise ValueSerializerError("Avro schema required for values")

        if key is not None:
            if key_schema:
                key = self._serializer.encode_record_with_schema(topic, key_schema, key, True)
            else:
                raise KeySerializerError("Avro schema required for key")

        super(AvroProducer, self).produce(topic, value, key, **kwargs)


class AvroConsumer(Consumer):
    """
    .. deprecated:: 2.0.2

    This class will be removed in a future version of the library.

    Kafka Consumer client which does avro schema decoding of messages.
    Handles message deserialization.

    Constructor arguments:

    :param dict config: Config parameters containing url for schema registry (``schema.registry.url``)
                        and the standard Kafka client configuration (``bootstrap.servers`` et.al)
    :param schema reader_key_schema: a reader schema for the message key
    :param schema reader_value_schema: a reader schema for the message value
    :raises ValueError: For invalid configurations
    """

    def __init__(self, config, schema_registry=None, reader_key_schema=None, reader_value_schema=None, **kwargs):
        warnings.warn(
            "AvroConsumer has been deprecated. Use AvroDeserializer instead.", category=DeprecationWarning, stacklevel=2
        )

        sr_conf = {
            key.replace("schema.registry.", ""): value
            for key, value in config.items()
            if key.startswith("schema.registry")
        }

        if sr_conf.get("basic.auth.credentials.source") == 'SASL_INHERIT':
            # Fallback to plural 'mechanisms' for backward compatibility
            sr_conf['sasl.mechanism'] = config.get('sasl.mechanism', config.get('sasl.mechanisms', ''))
            sr_conf['sasl.username'] = config.get('sasl.username', '')
            sr_conf['sasl.password'] = config.get('sasl.password', '')

        ap_conf = {key: value for key, value in config.items() if not key.startswith("schema.registry")}

        if schema_registry is None:
            schema_registry = CachedSchemaRegistryClient(sr_conf)
        elif sr_conf.get("url", None) is not None:
            raise ValueError("Cannot pass schema_registry along with schema.registry.url config")

        super(AvroConsumer, self).__init__(ap_conf, **kwargs)
        self._serializer = MessageSerializer(schema_registry, reader_key_schema, reader_value_schema)

    def poll(self, timeout=None):
        """
        This is an overriden method from confluent_kafka.Consumer class. This handles message
        deserialization using avro schema

        :param float timeout: Poll timeout in seconds (default: indefinite)
        :returns: message object with deserialized key and value as dict objects
        :rtype: Message
        """
        if timeout is None:
            timeout = -1
        message = super(AvroConsumer, self).poll(timeout)
        if message is None:
            return None

        if not message.error():
            try:
                if message.value() is not None:
                    decoded_value = self._serializer.decode_message(message.value(), is_key=False)
                    message.set_value(decoded_value)
                if message.key() is not None:
                    decoded_key = self._serializer.decode_message(message.key(), is_key=True)
                    message.set_key(decoded_key)
            except SerializerError as e:
                raise SerializerError(
                    "Message deserialization failed for message at {} [{}] offset {}: {}".format(
                        message.topic(), message.partition(), message.offset(), e
                    )
                )
        return message



# Copyright 2024 Confluent, Inc.
# Copyright 2023 Buf Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
import uuid
from email.utils import parseaddr
from ipaddress import IPv4Address, IPv6Address, ip_address
from urllib import parse as urlparse

import celpy
from celpy import celtypes

from confluent_kafka.schema_registry.rules.cel import string_format


def _validate_hostname(host):
    if not host:
        return False
    if len(host) > 253:
        return False

    if host[-1] == ".":
        host = host[:-1]

    all_digits = True
    for part in host.split("."):
        if len(part) == 0 or len(part) > 63:
            return False

        # Host names cannot begin or end with hyphens
        if part[0] == "-" or part[-1] == "-":
            return False
        all_digits = True
        for r in part:
            if (r < "A" or r > "Z") and (r < "a" or r > "z") and (r < "0" or r > "9") and r != "-":
                return False
            all_digits = all_digits and "0" <= r <= "9"
    return not all_digits


def validate_email(addr):
    parts = parseaddr(addr)
    if addr != parts[1]:
        return False

    addr = parts[1]
    if len(addr) > 254:
        return False

    parts = addr.split("@")
    if len(parts) != 2:
        return False
    if len(parts[0]) > 64:
        return False
    return _validate_hostname(parts[1])


def validate_host_and_port(string: str, *, port_required: bool) -> bool:
    if not string:
        return False

    split_idx = string.rfind(":")
    if string[0] == "[":
        end = string.find("]")
        after_end = end + 1
        if after_end == len(string):  # no port
            return not port_required and validate_ip(string[1:end], 6)
        if after_end == split_idx:  # port
            return validate_ip(string[1:end]) and validate_port(string[split_idx + 1 :])
        return False  # malformed

    if split_idx == -1:
        return not port_required and (_validate_hostname(string) or validate_ip(string, 4))

    host = string[:split_idx]
    port = string[split_idx + 1 :]
    return (_validate_hostname(host) or validate_ip(host, 4)) and validate_port(port)


def validate_port(val: str) -> bool:
    try:
        port = int(val)
        return port <= 65535
    except ValueError:
        return False


def validate_ip(val: typing.Union[str, bytes], version: typing.Optional[int] = None) -> bool:
    try:
        if version is None:
            ip_address(val)
        elif version == 4:
            IPv4Address(val)
        elif version == 6:
            IPv6Address(val)
        else:
            msg = "invalid argument, expected 4 or 6"
            raise celpy.CELEvalError(msg)
        return True
    except ValueError:
        return False


def validate_uuid(val: str) -> bool:
    try:
        uuid.UUID(val)
        return True
    except ValueError:
        return False


def is_ipv4(val: celtypes.Value) -> celpy.Result:
    if not isinstance(val, (celtypes.BytesType, celtypes.StringType)):
        msg = "invalid argument, expected string or bytes"
        raise celpy.CELEvalError(msg)
    return celtypes.BoolType(validate_ip(val, 4))


def is_ipv6(val: celtypes.Value) -> celpy.Result:
    if not isinstance(val, (celtypes.BytesType, celtypes.StringType)):
        msg = "invalid argument, expected string or bytes"
        raise celpy.CELEvalError(msg)
    return celtypes.BoolType(validate_ip(val, 6))


def is_email(string: celtypes.Value) -> celpy.Result:
    if not isinstance(string, celtypes.StringType):
        msg = "invalid argument, expected string"
        raise celpy.CELEvalError(msg)
    return celtypes.BoolType(validate_email(string))


def is_uri(string: celtypes.Value) -> celpy.Result:
    if not isinstance(string, celtypes.StringType):
        msg = "invalid argument, expected string"
        raise celpy.CELEvalError(msg)
    url = urlparse.urlparse(string)
    if not all([url.scheme, url.netloc, url.path]):
        return celtypes.BoolType(False)
    return celtypes.BoolType(True)


def is_uri_ref(string: celtypes.Value) -> celpy.Result:
    if not isinstance(string, celtypes.StringType):
        msg = "invalid argument, expected string"
        raise celpy.CELEvalError(msg)
    url = urlparse.urlparse(string)
    if not all([url.scheme, url.path]) and url.fragment:
        return celtypes.BoolType(False)
    return celtypes.BoolType(True)


def is_hostname(string: celtypes.Value) -> celpy.Result:
    if not isinstance(string, celtypes.StringType):
        msg = "invalid argument, expected string"
        raise celpy.CELEvalError(msg)
    return celtypes.BoolType(_validate_hostname(string))


def is_uuid(string: celtypes.Value) -> celpy.Result:
    if not isinstance(string, celtypes.StringType):
        msg = "invalid argument, expected string"
        raise celpy.CELEvalError(msg)
    return celtypes.BoolType(validate_uuid(string))


def make_extra_funcs(locale: str) -> typing.Dict[str, celpy.CELFunction]:
    string_fmt = string_format.StringFormat(locale)
    return {
        # Missing standard functions
        "format": string_fmt.format,
        # protovalidate specific functions
        "isIpv4": is_ipv4,
        "isIpv6": is_ipv6,
        "isEmail": is_email,
        "isUri": is_uri,
        "isUriRef": is_uri_ref,
        "isHostname": is_hostname,
        "isUuid": is_uuid,
    }


EXTRA_FUNCS = make_extra_funcs("en_US")
